;;;
;;; SokEvo and SokHard puzzles generated using SokEvo
;;; Author: Lee J Haywood
;;; Original website: http://fruise.googlepages.com/sokoban.html
;;; Collection: Haywood
;;;
;;; This collection of Sokoban puzzles was generated by a program written by
;;; Lee J Haywood that used random numbers and Brian Damgaard's YASGen program
;;; to 'evolve' them.
;;; The puzzles in the collection may be freely distributed and may be used
;;; commercially provided that they remain unchanged and I am credited with my
;;; name - Lee J Haywood.  Any similarity to anyone else's work is entirely
;;; coincidental and unintentional.
;;; The puzzles may be included with any program under the GNU General Public
;;; License, version 3.0 or later, with the following restrictions (see the
;;; Additional Terms section of the GPL).
;;; (1) A copy of this text must be included with the program, and also with
;;; all derivatives of that program which include any of the puzzles.
;;; (2) Misrepresentation of the origin of the puzzles is prohibited, and
;;; modified versions must be marked as such.
;;;
;;; Author's notes:
;;;
;;; The SokEvo puzzles are sorted by the length of their shortest solutions,
;;; which is intended to be an objective way of ordering them from easiest to
;;; hardest.  Your abilities are likely to mean that some seem to be out of
;;; place, however.  Also, some puzzles have more solutions than others so may
;;; be easier as a result. The later puzzles are extremely difficult.
;;;
;;; The SokHard puzzles are sorted in release order.
;;;

(in-package :cl-sokoban)

(defun add-solution (&rest lines)
  (push (x:join lines "") *solutions*))

;;; mazes & solutions

(defmaze
  "#######"
  "#.@ # #"
  "#$* $ #"
  "#   $ #"
  "# ..  #"
  "#  *  #"
  "#######")

(add-solution
  "DDrdrruLruLLDllU")

(defmaze
  "#######"
  "#.    #"
  "#$* # #"
  "#.  $*#"
  "# .$  #"
  "#@ *  #"
  "#######")

(add-solution
  "uuUdrruurrdDLLLrrddrUlL")

(defmaze
  "#######"
  "#     #"
  "#@$.# #"
  "#*$  .#"
  "# $$  #"
  "# . . #"
  "#######")

(add-solution
  "RlDRRRddlllUdrrruullDRurD")

(defmaze
  "#######"
  "#   ###"
  "#  ** #"
  "#   $ #"
  "#  $..#"
  "###@  #"
  "#######")

(add-solution
  "ruLulldRRdrruuLLdRllluurDldRR")

(defmaze
  "########"
  "# *   .#"
  "#  $ * #"
  "# .$ $ #"
  "#* *  *#"
  "#.# #  #"
  "#    @ #"
  "########")

(add-solution
  "uuluLruulDlldDuurrrRlddLLrruulDrdrUrU")

(defmaze
  "#######"
  "#  .  #"
  "# $*$$#"
  "# ..$.#"
  "#@* $ #"
  "##.   #"
  "#######")

(add-solution
  "urDrdrruuLrddllulluuurrrrDulllDlddrrdrruLLruUruL")

(defmaze
  "########"
  "#  #  *#"
  "#  * * #"
  "# *$ $ #"
  "#  .#. #"
  "#   # *#"
  "## @   #"
  "########")

(add-solution
  "rruuruuLDDuuulDLDRurrddlUllDlluuRlddrddrUdrrU")

(defmaze
  "#######"
  "#     #"
  "#$* # #"
  "#.@ $.#"
  "# .$$$#"
  "#  .. #"
  "#######")

(add-solution
  "ddrrrUULLDuuullDDRluurrdDlddlUrurrDLuuurrD")

(defmaze
  "########"
  "#   *@ #"
  "## .. .#"
  "# *$$.$#"
  "#  $.$ #"
  "# $ *.##"
  "##   ###"
  "########")

(add-solution
  "dlluRdrdDlUruLulldRlDlddRdrrUUruuruLdddrUdllddlluRlURUR")

(defmaze
  "#######"
  "#  $. #"
  "#@*$ $#"
  "#.* . #"
  "# $  .#"
  "###   #"
  "#######")

(add-solution
  "ddRRuruurDDllddrUllluuurRDDLdRuurrdLLddrU")

(defmaze
  "#######"
  "#  .$ #"
  "#. @* #"
  "# $   #"
  "#$**$.#"
  "#..$  #"
  "#######")

(add-solution
  "drruLLullddDuRRuLrrrdddlLUUdRUlluurDDrruuL")

(defmaze
  "########"
  "#.  ## #"
  "#  $   #"
  "#*$ $. #"
  "# $.#..#"
  "#@$  $ #"
  "# #  *.#"
  "########")

(add-solution
  "uUUdddRRuuLrRurrddddLUruuulldlddRRlldRlulluRuuurDrrrddDuuulldRurDllull"
  "dRR")

(defmaze
  "#######"
  "# *   #"
  "# $  .#"
  "#.# * #"
  "#@.$$ #"
  "# $ . #"
  "#######")

(add-solution
  "dRRUlluuRRurrdddLLLdlUrrrruuulldDuurrdddlLuuR")

(defmaze
  "#######"
  "#  ##.#"
  "# $$ $#"
  "#. .. #"
  "#$ *  #"
  "#@*   #"
  "#######")

(add-solution
  "RuRurrddlUlluluurDldRdrrruUlLDlluurDldRddlUruurrrdddlL")

(defmaze
  "#######"
  "#   . #"
  "# .@$$#"
  "# * * #"
  "# $  .#"
  "###.$ #"
  "#######")

(add-solution
  "urrDLLddrrULululldddRRUrrddLUluUrurDDullddlluuurR")

(defmaze
  "#######"
  "##    #"
  "#..$ @#"
  "# $ **#"
  "#$*$  #"
  "#.  . #"
  "#######")

(add-solution
  "lulldRdDuulldDRUruurrdDLddrUluLLddRlulUrrrdLuuurD")

(defmaze
  "########"
  "# *   .#"
  "#  *   #"
  "# . #$ #"
  "#   $. #"
  "#@* #$ #"
  "## $ . #"
  "########")

(add-solution
  "RdRUUluuRurrdLLrrrddlUdLLdlluurDluuuRDrrRdrdddlUruuUlluLdllddrrUdddR")

(defmaze
  "#######"
  "# @   #"
  "#$* **#"
  "#. #$ #"
  "# . $ #"
  "#  .  #"
  "#######")

(add-solution
  "rrrDLullDDlddrrrruUdLdllluurDldRuuuulDrurrdLrDrddlUUdL")

(defmaze
  "########"
  "# ..@$.#"
  "#. $   #"
  "#  $ $ #"
  "#.# * ##"
  "# #$$# #"
  "#. $ . #"
  "########")

(add-solution
  "ddLUllddddrRUUUruuLrRdrdLdLUlUllddddrrRUdllluuuurrdLulDrrrruLLLrdDDrdL"
  "L")

(defmaze
  "#######"
  "#.    #"
  "#.$$$ #"
  "#. * .#"
  "#@$$*$#"
  "#. $..#"
  "#######")

(add-solution
  "uuurrrrddDuuullllddrRRUdllluurrDrdLuullddddrRUULrruruLLLrDLdDrdL")

(defmaze
  "########"
  "### *  #"
  "#   $ @#"
  "#   #. #"
  "#*  $* #"
  "# * .* #"
  "##  *  #"
  "########")

(add-solution
  "lLLddddlUruuurrrddddlLuluulluRRRldddrdrruLruulDruuulDllddRldldRuuulD")

(defmaze
  "########"
  "#. $   #"
  "#.$##$ #"
  "#$ $.  #"
  "# $#.$.#"
  "#.   # #"
  "###  @.#"
  "########")

(add-solution
  "llulluUrDldRRdrrruuuuullLLDDDuuurrrrdddddllluLrdrrruuulldRurDDuuuulDDl"
  "dRuuulllddRlddrdrU")

(defmaze
  "########"
  "#  .## #"
  "# $   @#"
  "#  **$ #"
  "#*.. # #"
  "#  * $*#"
  "# *    #"
  "########")

(add-solution
  "lllullddrdrUUddrddlUUrddrrULLdlUluuluurDrrDDLddrruLuuurrdDuLullDulldR")

(defmaze
  "########"
  "#  .   #"
  "# $##$ #"
  "#*$   .#"
  "#  *$@ #"
  "##$. # #"
  "#  . . #"
  "########")

(add-solution
  "ulDDldRuuulDllURRRdddllUdrruulDuulldRurDrrruuullllDDldRuuulDrdrrRUruLL")

(defmaze
  "#######"
  "#   @ #"
  "#$* * #"
  "#. *$ #"
  "# . $##"
  "#  .  #"
  "#######")

(add-solution
  "ldDuurrdLrdLuulllDDrddrrUUddlluuRDluluurrDDlddlUruruurrddLdL")

(defmaze
  "########"
  "#. .$ .#"
  "# $  $$#"
  "#@ $## #"
  "# *$ $ #"
  "#.* .. #"
  "##  *. #"
  "########")

(add-solution
  "ddRUddrRuuLDldRluluuuurrdDDLruuullddDrrrdrruuULLrrdddllulldRRlluluuurr"
  "DDDldRuuuLDDluUrrrruLrdrddLLL")

(defmaze
  "#######"
  "#   #@#"
  "#  $$ #"
  "# # $ #"
  "# . ..#"
  "# #   #"
  "#######")

(add-solution
  "dddlllluururDDullddrRdrruuuLrdddlluRllluururDDullddrRdrrULLuurDurD")

(defmaze
  "########"
  "#     ##"
  "# .*.$##"
  "# * .  #"
  "#* . $##"
  "#.$$ $ #"
  "#@$  . #"
  "########")

(add-solution
  "uUrUdlddRUUrrddrruLdlULuurrDLuuullldDuurrrrDDldddrruLdlUUUrDlddrruLdlU"
  "UdlLdRR")

(defmaze
  "########"
  "#   *@ #"
  "#. #$. #"
  "# *$ * #"
  "#$ . $ #"
  "# .$# *#"
  "#   $..#"
  "########")

(add-solution
  "rdddDldLLUluURDrRUrDlddllUluurRdLulluurrRDDldddrruuLrddllulldRRRllluUU"
  "rDurruullDurrdddrruuuL")

(defmaze
  "#######"
  "#   *.#"
  "# *$$ #"
  "#. @#*#"
  "# . $ #"
  "# $.  #"
  "#######")

(add-solution
  "lluurrDLddrdrruUULLulDDurrrdddlluUddrruuulLddRdrUllullddRluurruullD")

(defmaze
  "########"
  "#    * #"
  "#.#$$#.#"
  "#   *@ #"
  "#$*. *$#"
  "#. .$$.#"
  "#    $.#"
  "########")

(add-solution
  "ruuLLLDDlDurRdLuuurrrddDLruuulllddlluuRRdddddlluUUrrrRdrUdlLuluuRDldll"
  "dddrruLrdrRUL")

(defmaze
  "########"
  "##  ...#"
  "#.$$* .#"
  "##$.$$ #"
  "# *  $##"
  "#@  $*.#"
  "# $  *.#"
  "########")

(add-solution
  "rruuULuRdddddrRllulluRUruuRRdLullDDDlddRRuUrUUddlluuurRdddlddRURlUUUrr"
  "dLDllluuRRRddlUdddllU")

(defmaze
  "########"
  "#@#  . #"
  "#.$$#. #"
  "#  $  *#"
  "# #$$. #"
  "#$ $  .#"
  "#    ..#"
  "########")

(add-solution
  "ddrRRDDLdRlllURRRuulDDrurrULuurDDDluuullDDlluRdrRRuullDldRRdrUUddldlUd"
  "llUUrRRdrUdlddRUllldRR")

(defmaze
  "#######"
  "#.    #"
  "# $$$@#"
  "#. # .#"
  "# $$*.#"
  "# $ ..#"
  "#######")

(add-solution
  "ullllddddRRRllluuuurrrrddlDruuullllddrUdlddrURRuruulDDuulLdRRurDlllDld"
  "drruLdlU")

(defmaze
  "#######"
  "#  ##.#"
  "# $$ $#"
  "#. .  #"
  "#$$*  #"
  "#@. .##"
  "#######")

(add-solution
  "rrruullluurDRdrddlllUURRDullddrRuuluulDDrrdrruULLdlluurDldRurrrddlldll"
  "UdrrurruulDrdLuulldD")

(defmaze
  "########"
  "#  $. ##"
  "#* # $ #"
  "#.*.   #"
  "##.$##@#"
  "# $$ $ #"
  "#..$.  #"
  "########")

(add-solution
  "ulluurDldlLuuRlddrrurrddddllLLUUddrruLdlUrrdrruuuulldlDDldRRluuururrdd"
  "dLLrruuulldldDrdLrurrdLruuuLL")

(defmaze
  "########"
  "# @.$ .#"
  "# $   ##"
  "#.*$#  #"
  "#$ $.$ #"
  "#.* . ##"
  "##  *  #"
  "########")

(add-solution
  "rRRllllddDrRUUdddrrULdlUlluuurRdrrDrdLdlldlUrrruuullDDuuulldddRdRUlluu"
  "urrdrrddLLrruullDLDRuuulDrrruL")

(defmaze
  "########"
  "#  #####"
  "#  .####"
  "# $$$  #"
  "#.$  $.#"
  "#$#  #.#"
  "#@. * .#"
  "########")

(add-solution
  "rrRRluuRurDlllURdddlllUURlddrrruuulDllURuulDDDrrruLrdrruLdlllluurDldRR"
  "RRuLLdlluurDldRurrdddlllUdrrruuulldRDrdLuuurDD")

(defmaze
  "########"
  "#  . $ #"
  "# $ .  #"
  "#  $#$@#"
  "#*# * .#"
  "#   .$##"
  "# * .$.#"
  "########")

(add-solution
  "ulllDDRddlUllUUrrDuurrrddLLrruullldllddrRdrRUdllulluurrDrDulullddrRuuu"
  "ullDDrrurrrddLUruuLLdRurDDlulldlluRR")

(defmaze
  "########"
  "# .    #"
  "#$$ @$ #"
  "#.$# $ #"
  "# $.#  #"
  "# .$ **#"
  "#. ..$.#"
  "########")

(add-solution
  "urrdLrddlUlUlullDDDRddlUrdrrRUUruuulLdlLDDDldRRluuuurrurrdddldlLrruruu"
  "ulldlluRdlddDldRuuulDDrrrruruuLLLulDDDuuulDrrrrDDuuuLL")

(defmaze
  "########"
  "##  @  #"
  "#.#$##.#"
  "#$  $  #"
  "# #.**$#"
  "#  *$ .#"
  "#.   $.#"
  "########")

(add-solution
  "rrddlLLdDRdLuuurDurrDLulldddrRUULrddlluRuulDrrruuulllDDrdddrUdllLuRRuu"
  "lDuuurrrddLLrruullldDrdddlluluURRuurrrdddLruuulllddRRdrUlllllddrdrrrU")

(defmaze
  "########"
  "#   ## #"
  "# $ @  #"
  "# #$#..#"
  "# $ #..#"
  "# $  * #"
  "# #  * #"
  "########")

(add-solution
  "rrddddLUUruulllullddddRRRllluuuurrdrrrddddlUruuulllDDuuulldddRdRRRdrUU"
  "dllllluuuurrddDldRRRdrUllllluuuRurDDDldRRldR")

(defmaze
  "#######"
  "#.  $@#"
  "#$$ $$#"
  "#. .$.#"
  "#.$*$.#"
  "## . .#"
  "#######")

(add-solution
  "DDLLDLdRuuuullDRDrrruuLLDldRRlddlUrdRUluuurrdLulDDDrUluurrdLulDlluRRdr"
  "ddllUdrruulDuulldRurDrruLLL")

(defmaze
  "#######"
  "##. *@#"
  "#..$*$#"
  "# $   #"
  "#$ $*.#"
  "# .   #"
  "#######")

(add-solution
  "LLDurrDLLrdDrUdlLdllURRUruulDDrrddLUUluurrDullddrddLUlldRRuruUlDrdLdll"
  "uURRdrrdLLuuurrDullddrUllldRR")

(defmaze
  "#######"
  "#.. * #"
  "#.#$$ #"
  "#$  $ #"
  "# .$*.#"
  "##. $@#"
  "#######")

(add-solution
  "uuuuLLLrrrddLLUdrrddLLUUdllURRddrruLLrruuulLDDRluurrdDlddlUUlldRRurruu"
  "llDurrddlUdLdlluRRddrrULLuurDurD")

(defmaze
  "########"
  "##    .#"
  "## $$*.#"
  "#. * $$#"
  "#* .#  #"
  "#@$$$ ##"
  "#. .* .#"
  "########")

(add-solution
  "drrRRuurUUddlddlllluUruuurrDRlulldRlddrUluurrrdDLLdllddrrrruUruLuullld"
  "DrdLDuuuurrrdddddllLuurDldRluuuuurrrddddLruuuullDRllddrDluuurDlddrrruu"
  "L")

(defmaze
  "########"
  "# #    #"
  "# $@$  #"
  "# $#* $#"
  "#.$  *.#"
  "# $ #  #"
  "#.... ##"
  "########")

(add-solution
  "urrddDllddrrUruLuuulldRurrdDldddlllluuuuRRllddddrruuLrddlluRdrrruuuruu"
  "llldlDDDldRRluuulDDruuururrrddldLLDlluurDDuuururrDulldRDuurrddLdLLLuur"
  "rurDD")

(defmaze
  "########"
  "#      #"
  "#$@*$$##"
  "# $ $ ##"
  "# $.#..#"
  "# ..*$.#"
  "#.     #"
  "########")

(add-solution
  "ulDDDDrrdrrruuluLrdrddlllluluuururrDLullddddrruUddlluuRDluuurDRllddrUd"
  "rUdddrrruuluLLdddrrUdlluRRdrUllluurrDullddrRllulldRRlluuuurrDDDuuurrDD"
  "uLulDDuulldRDD")

(defmaze
  "#######"
  "###  .#"
  "# $$# #"
  "# @$. #"
  "#. $# #"
  "#   . #"
  "#######")

(add-solution
  "ddrrruulLrruullDurrddddlllluuuRDlddrrrruulLLrrruullDldRuurrddddlllUdrr"
  "ruuuullddRlullDRddrrruUUdddllluuRDldR")

(defmaze
  "########"
  "##  #  #"
  "# $  $ #"
  "## **  #"
  "# $    #"
  "# # #$$#"
  "#...@..#"
  "########")

(add-solution
  "luurUrrdDulllddlluuRRluRuulDDrdddlluuRurDllddrrrrUruulLrrddldlllluurrD"
  "uluurDrrrddldlLLrrruruulllDDuurrrddldlLuuurrDDuLrruuulDLdlDDuluuurDrrr"
  "dL")

(defmaze
  "########"
  "#     ##"
  "#.#.* .#"
  "# $$$  #"
  "#*. $* #"
  "#    #@#"
  "########")

(add-solution
  "uuluulldDRluullddRddrrULUlluurrDurrdLrrddLLURdrUluullllddrRRllluurrrrd"
  "DlddlllUUddrrULrrurrdLLdlluuRlddrruLrruuullDDRluurrdDlddlUU")

(defmaze
  "########"
  "#     ##"
  "#@$$ $##"
  "#   $ ##"
  "# *## .#"
  "# ..*$.#"
  "#.#    #"
  "########")

(add-solution
  "dddrrdrrruuluLrdrddlUdlluRRlllluuuurrrddrDuluulllddddrrrdrrULLLLrrruul"
  "uurDDDuluullldddDrrrRlllluurDldRRlluuuurDDDldRuuuurrdLulDDDuurrdLulD")

(defmaze
  "########"
  "# . $ ##"
  "## $.$ #"
  "#...# .#"
  "# #$ $ #"
  "# $ #.$#"
  "# @    #"
  "########")

(add-solution
  "luuurrDulldddrrrruUruuLDDlluluurDldldddrrrrrUUluuuLrdddrddllllluuururR"
  "urDrDDlllUluuRlddldddrrUUrrruullLulDrrruLLrrdrddlUruLddlldLdlUUdrrdrrr"
  "UU")

(defmaze
  "########"
  "## * .##"
  "#   .$ #"
  "#.# $  #"
  "#. *$# #"
  "# $#   #"
  "#  ##@##"
  "########")

(add-solution
  "uruulLulluRRddrrddllULUdrdrruulLuulldlddRRlluururrdLrdrrddllUlURuulDrd"
  "dlllddrUluuuRurrdddlLdlUrrruuulldRDrdLLruuurDlddrdrruuuLLDlUrrrdLLulDr"
  "rrddllUU")

(defmaze
  "########"
  "#...   #"
  "# #.## #"
  "# #    #"
  "# $  *$#"
  "## #$..#"
  "# $ $$ #"
  "# @ $ .#"
  "########")

(add-solution
  "rRRULUUluuurrrddlLrrDLDlUdddllURdrrUrULuurDlllURdrruuullllldddRDuluuur"
  "rrrrdddllulUdrdrruuullLLrrrrddLLrddddlluRUUdddllluRuuRUUddldddrruururu"
  "uullLrrrddlLrddlUlUUddrruLddddrUUlulUdlddRRdrU")

(defmaze
  "########"
  "#  ..# #"
  "##$$   #"
  "# @ $..#"
  "# #$*$ #"
  "#.* $. #"
  "# #    #"
  "########")

(add-solution
  "rRurrddddllluRdrruuuulldlDururrddddlluURUrDllluRdddlUruulllddRRUruLruu"
  "lDrddrruuLDldlddrrrUULulUlDrrdrddlllulluuRRuulDrrrddrddllluUUUdddlluuR"
  "urDDuurrddLrrddllULLrrdrruuluulldDuurrddrddlluLuuR")

(defmaze
  "########"
  "# @* #.#"
  "# $$ $ #"
  "#  # #*#"
  "##  .  #"
  "#  $.# #"
  "##  * .#"
  "########")

(add-solution
  "lddrdddrRlluuuluurDDDrruuuLrdddlluuRurDllddrrrrUULLDurrddlLddllUURRuur"
  "rddddLLuuRlddrruUlllldRldRRRllluuUluuRDDDrruuuLrdddDluluuRurDDDlddlUUU"
  "luRRurDDullluR")

(defmaze
  "########"
  "#  .$  #"
  "# *$$. #"
  "#    $##"
  "#*$. $.#"
  "#  .#$ #"
  "##@.. ##"
  "########")

(add-solution
  "ruuulluurDldRDrddlUlUUrDuuurRDLulDrrrruLLdllDlddRUrrRUdllURdllluRuulDD"
  "rdrruuLrurrdLLddrUluurrdLulDrddlllddrrrUUllDldRluuuluurDRDDDuuulldRurR"
  "urrdLLLDDrrULulD")

(defmaze
  "########"
  "# #    #"
  "# @*.* #"
  "# #$$$ #"
  "#.  * .#"
  "#.#$$$ #"
  "#  . . #"
  "########")

(add-solution
  "lddddrrrrruuuuulldLrurrdLrddddllllluurrRRUrDlllURdrruuullDDldRuuulDrdd"
  "llluuRRDrdLuullddddrrrUdllluuuurrddLruuurrrddLLrruullldlldDrrDrUllluur"
  "rDuurrrddlLdLLrUrrrdLullddrdrrUdlluluurrdD")

(defmaze
  "########"
  "# $.  .#"
  "# .$.  #"
  "# # $#.#"
  "#@$* $ #"
  "#.$ #$##"
  "#.     #"
  "########")

(add-solution
  "uuuRRRdLulldddddrruUUULulDDDDuuurrddLruuruLrdrruLrdddLLLddrrUdlluuuull"
  "dDRluurrrrrddLddlluUUdlluurRRRurDlllddddlUruuullddRdrUUdlluurRuRRdLLdd"
  "lluuuRldddrrddrruuLU")

(defmaze
  "#########"
  "##  $. .#"
  "#  ## #$#"
  "# $#. ..#"
  "#  @## .#"
  "#* $$   #"
  "#   # $ #"
  "###   # #"
  "#########")

(add-solution
  "lddrdrruuruuluurrDDlddlddlluluurDluluururRRddrddlLLrrddllUllUUrDRddrru"
  "urrdLuuurDluluullldDDrdRRllddrrUrUUruuuLrdddlddrUUUUdLLrdddldlluurRdrU"
  "UdlllullddRluRRRRRdrUUddlldllUluuuulDDrdRRRRdrU")

(defmaze
  "########"
  "##  .  #"
  "# #$@$ #"
  "# $. *$#"
  "#  *.$ #"
  "# . #$.#"
  "# #.   #"
  "########")

(add-solution
  "ddLruuurrdLrDDLUlldRurruulllDurrrddllUdrruulDlddlddrrrUULrULuurDllluRd"
  "rddrddllluuUdddrrruulLruulDrdrddlllulluuRRRurrDDlUllllddrrdrrrUdlllull"
  "uurDurruulDrdrdDrdLLruuurDuuuLD")

(defmaze
  "########"
  "# # . .#"
  "# $ #  #"
  "#. $# $#"
  "# $@*. #"
  "# #$  .#"
  "#  $.###"
  "########")

(add-solution
  "RdrruUlDrdLullUlldRRRRuuullDDuurrdddlLulldddrRUULrrruuulldDuurrdddlLul"
  "luRdrdrruuullDurrddrDlllddlluURuururrrDldddLruuuulldlddRRdrUUUruLdddrU"
  "UddlllluurDldRRlllddrrUdlluurR")

(defmaze
  "########"
  "#   . .#"
  "#@$#$  #"
  "# .$ $ #"
  "# $ #  #"
  "##*$# ##"
  "##... ##"
  "########")

(add-solution
  "urrrrrdddlUruulllllddrRRRdrUUdllllluurrrrDrddlddllUUddrruuruulullllddd"
  "RluuurrrrdrddlddlllUrdrruuruulullllddrRDDldRuuulldRluuurDDDDuuuurrrdrd"
  "dlULLDDluluurDDuuurrrDrdLLLrU")

(defmaze
  "#######"
  "#    .#"
  "#@# $ #"
  "#*  $##"
  "#     #"
  "#$$##.#"
  "#. *. #"
  "#.  $ #"
  "#######")

(add-solution
  "urrdddrrddlLdlluRUUlDruruuurrdLullldDRdrrUUddlluluurrRddLdldlUrdddlUru"
  "urruuulDDrdrdddLLLuuulURddddrrruuulLUlldRddRRdrUllllUruurrdLulDDldRRlu"
  "urruuulDDrdLulDDldRuuulDDruruuullD")

(defmaze
  "########"
  "#    # #"
  "#.# $* #"
  "#$$ * @#"
  "#. .$.##"
  "#  #$.##"
  "##     #"
  "########")

(add-solution
  "ldddllluuruuulldDRdrUdlddrrruuuLrdddllluurRlluluurrrDDrDLuuulllddrdddr"
  "rUUUdddlluuuRldddrruuLulDrrruLdlluluurrrDDuulllddrdrRddllUlURddrruuuuu"
  "llldDrRlluurrrdDllddlUUrrrDLLrruuulDDldddrrUU")

(defmaze
  "########"
  "##@ .. #"
  "# #$#. #"
  "#   ..*#"
  "##$$ $ #"
  "#  $ $ #"
  "#   # .#"
  "########")

(add-solution
  "rrrddlddRUdlLdlluRRRuulDrrrULuullDDRluurrdDlddLUUdddlluRdrUruullDRddlU"
  "ruurrrdDLLLuurruurDDlllddrrUUlldRdldlluRdrUruullDurrrruulllDurrrddlUru"
  "LddlddrrUdlUUUdddlldlluRuRRdrUUdllluRRdllddrUUluR")

(defmaze
  "########"
  "#@$. $.#"
  "#  $ $.#"
  "# $$ $ #"
  "#. $.#*#"
  "###$*. #"
  "# . . .#"
  "########")

(add-solution
  "dddrRDRdLuulluurRDrDLddrrruUULLdDldRuuurrddLLulLURRluurDlddrddlUUluluu"
  "RDrrDDlUdddrUrruuLLuuRDldDllURdrUdlddrUrrdLruuuLLDldRRdrUlluuruulDDDld"
  "RldRRlluululuurDRDrruulDDDDuululldRRurDDulldR")

(defmaze
  "########"
  "## ##  #"
  "# $@$$ #"
  "#.#    #"
  "#.#$## #"
  "#   $. #"
  "# *.  .#"
  "########")

(add-solution
  "drrrdddlllulldRRRRuruulllDDldRuuurrruulDrdddlLLuuuRdlddrrruuLLrrddllld"
  "lluuuuRRDrrrddlllLdlUUdrrrrruulllDurrruulDrdddllLLruuuRdlddrrruuLLulDD"
  "urrrddllLdllURRuurrrddldLLrruruuuulDrdLLulDDDR")

(defmaze
  "########"
  "##   # #"
  "#  $$ .#"
  "# $@$###"
  "#.#$   #"
  "#.* ## #"
  "#.  .  #"
  "########")

(add-solution
  "DrrrddllllluuuururDRRllllddddrrrrruulllUdrrrddllllluuuururrdDuulldlddd"
  "drrUURuuLrddlUdddrrruuLLulDrrrddllllluuuRldddrrrrruulllDuruuullDlddddr"
  "RuuruuLrddlULrruullDlDDDuuururrddlLulDDuruurDrdLLulD")

(defmaze
  "########"
  "# $. ..#"
  "#.$.$$ #"
  "#  .$  #"
  "###  ###"
  "#..* $ #"
  "#$$*$  #"
  "#@*  .##"
  "########")

(add-solution
  "RUdRRULUUrDluUlluuRRDuRRDLulllddrUluRRRdLrDLddrUUluullddRluurrdDrddlUU"
  "lluurrrDDurrdLLLddrUdlddllURdrrUUlDrdLuuuuRuulDDDDrUUluullddRluurrdDrd"
  "dlUUlluurrrDDurrdLLLddrUluRuulDDDuuullddRluurrrdddlUULrrrrdLLdddrruLdl"
  "UUUruL")

(defmaze
  "########"
  "# #@   #"
  "#. ** ##"
  "#. #   #"
  "# $$ #.#"
  "# $ ** #"
  "#  .$  #"
  "########")

(add-solution
  "rrddrdddlLUlLdlUrUdrrdrruuullDLruruulDLrrdrdddllLuurUdldllURdrdrrruuul"
  "uulldLDDldRuuururrddldLruruulldldDrruruLrdrdddllURdrUllUlllddRURRdLruu"
  "UruulDDDldlluRuuRurrddldLruruulldldDrruruLddlddrUUUdllldRRlldR")

(defmaze
  "########"
  "#.  . ##"
  "# #  $ #"
  "#$*$##@#"
  "#  $.# #"
  "#. # $ #"
  "##  .  #"
  "########")

(add-solution
  "dddlluuLrddrruuuuLulldDuulldDRDuluurrrrdrddddlluuLUUddrddrruuuulLrrddd"
  "dlluulullDRluuurRDrrrddddllllUUddrruuLrddrruuuulllDLrurrrddddlllluUlUU"
  "dRdddrruuLrddrruLdllluUrrDullddrRurruuulllDurrrdddlldlluuRlddrrurrdL")

(defmaze
  "########"
  "#  ..# #"
  "# $  $ #"
  "#*$#$  #"
  "# @$.*##"
  "# #.   #"
  "#  . ###"
  "########")

(add-solution
  "lddrrruuUUlulldDDrRlluuurrdrdDllluuRlddrUUddrruuLrdrruLLddlddrUrUUlDrd"
  "LullluuuRDDldRRRuuLrdrruLLdddrUlllluuurDDldRRRuuLrddllluuurDDurrddddll"
  "lUUUrurrrrdLullldldddrrrurUdldllluuururrDDlDurrdLuuurDlulldldRRluuulD")

(defmaze
  "########"
  "#   .$ #"
  "#$@* #.#"
  "# **$  #"
  "#  $.$$#"
  "#.#. #.#"
  "# $   .#"
  "########")

(add-solution
  "DldddRRRRluuULDrddllluuRluRuulDrddlddrrruuuLrrruuLLLDllDDrURRdLuuurrrd"
  "dDLruuulllddrRdrUllluurDlddRddlUUUlluuRRDrDDuLdddllUURlddrruUrurrdLLuu"
  "lulldRlDDrURRllluurrDrdLrdLddllUURlddrruUruulDDrrruLLdlllddrrrUdllluur"
  "rRurrdLulululldRlDD")

(defmaze
  "########"
  "#@$.   #"
  "#$*$  .#"
  "#.$ *#.#"
  "#  $ $ #"
  "# #$..##"
  "#.. $  #"
  "########")

(add-solution
  "DDRluuRDRRDDLUruurrdddLddLLURUUlDrrruuullLDlldDDuRUdRRuLuurDlddlluuuRl"
  "dddrruuLrddlluRurrurrdddLLLUluuRRdLddrrruuuLrdddllluurRurDlllddrUllldR"
  "uruullDRddlUrrrdLuluurDldRdrddlUUUlldRurDluuulDrurDRRuLLdlddrUdrddlUUl"
  "uuRurrdLDDlluuluRdddrruuLrddlluR")

(defmaze
  "#######"
  "###.###"
  "# . ###"
  "#.$*$.#"
  "#..  .#"
  "#*#$$$#"
  "#  @$.#"
  "##$$  #"
  "#     #"
  "#######")

(add-solution
  "llUUrrURdrDDLLURdrUluulDllddRRUruulUllDRdRddlluUrrdrrdLLuuluurDrDulldR"
  "ddrddlUUUUllddRluurrdDrddllUdrruuluullddRddrruuruLdddlluuRlddrruLulluu"
  "rrDrDLuuluurDrDDlUddddrrULdlUUUrDuuuLDllddRRddlUruUruulDllddRdrdrruLdl"
  "luuluurrrdDuulllddrRUruulullDDrUrrdLLrddlddrUUUUruLLrdddddrruLdlUUUUrd"
  "rDLddrUllUUruL")

(defmaze
  "########"
  "## * *.#"
  "#..$ $*#"
  "# .# *.#"
  "## $ $##"
  "# #$ $ #"
  "#..  $@#"
  "########")

(add-solution
  "LLLLrruuuuLLddRDrdrruLLUlluurrDRDLuuuRDDldddrruLdlUUUrDuuulDllddRRddLU"
  "ruUruulDllddRdrdrruLdlluuluuuRdrrdDuulllddrRUruLLulDrrruLdddlddrUUUddd"
  "rruLUUULLrddldRdrUU")

(defmaze
  "########"
  "#  . $@#"
  "#.#$*$$#"
  "#  . * #"
  "#$$ $$.#"
  "#.# #  #"
  "#.  . .#"
  "########")

(add-solution
  "DDLDDrdLuuurDuuuLLLDDRRlluurDurrdLddddLLLruUURllluuRRRDrDLLdddrruULulD"
  "ruuulDrrruLdlddrrULullullddDRRRUdllluuurrdDLdlUrrrruuLrddldLuuRdrrdLdd"
  "llUURlddrruUluulDDrrruLdllulldRRuururrDLdLLuRdrdddllUURlddrruUluulDDll"
  "uRRdrrddllUdrruulLulldRRuururrdLDDrUlLLdddrrrUdllluuuuRdrdDrdLuuulldll"
  "uRdrRurrDDlUluuurrDDuL")

(defmaze
  "########"
  "#  . $@#"
  "#.#$*$$#"
  "#  .$. #"
  "#$$  $ #"
  "#..$# .#"
  "##. * .#"
  "########")

(add-solution
  "LLLDDlDurRdLuuurrrDDDLruuulllddrRlluurDurrdLddddLLUUURllluuRRRDrDLLddd"
  "rruULulDruuulDrrruLdlddrrULullullddDRRRUdllluuurrdDLdlUrrrruuLrddldLuu"
  "RdrrdLddllUURlddrruUluulDDrrruLdllulldRRuururrDLdLLuRdrdddllUURlddrruU"
  "luulDDlluRRdrrddllUdrruulLulldRRuururrdLDDrUlLLdddrrrUdllluuuuRdrdDrdL"
  "uuulldlluRdrRurrDDlUluuurrDLrDD")

(defmaze
  "########"
  "#    . #"
  "# # $* #"
  "#.. .$ #"
  "# $$.$ #"
  "#..$#$*#"
  "#    *@#"
  "########")

(add-solution
  "UUULLLLDlUrrrdLrurrdLulluurrrDulllddrrUdlluurDrddrddLLLUUURurrDDlUlldd"
  "drrUULullldddRRRluUruLruulDrddrruLruuLDllullDDDRRRUdllluuurrdDrdLddllU"
  "UURldddrruuLulDrrruLuullDurrRDrrddLU")

(defmaze
  "#########"
  "# #.#   #"
  "#   $$$.#"
  "#$$$$   #"
  "#  $@#  #"
  "#. ###$##"
  "#..... .#"
  "#########")

(add-solution
  "LUUllDRDDldRRRRRUUruuullDLLrrurrdLrddlddlllluuulDrrrULdlddrrrruuULrruu"
  "llDRddddlllluuruuRRurDlllddrUluRRurrDLDLLrruulDlddllluuRRRddllddlUUrrr"
  "uulDruRurrdLdLLrruulDlllDDlddrrrrruUUdddllllluurDldRRRlluuuurrdLulDDDl"
  "dRRluuurrururrdLdLLLulDDDldRuuuulDDDruurrdLrurrdrUluulDrdLLLLulDDurDD")

(defmaze
  "#########"
  "#..  .$@#"
  "#.##$#*$#"
  "#  # .* #"
  "#$* .$$ #"
  "#   $ $.#"
  "# ##  # #"
  "#  #  #.#"
  "#########")

(add-solution
  "LLLLLrrrrrDDLUrdDDDuLULLDRurUruuLLLLrDDRRdrUllluurrDDlddlUUrddddlUUruu"
  "rrddLruuluurDDlldDlddrUUrUUllDDrUdllllUUrDldRRRuLrrruruulllDDRdrdrUllL"
  "dRddlUUrurrdLululuurrrdDDllDlddrUUruruuulllddDDrruruLUruLddddlluRdrUUd"
  "dllllluurDRRRdddlUUlulldRRuRDrrULLLrrdddlUU")

(defmaze
  "##########"
  "## . #. ##"
  "# .*...# #"
  "#$ $  $# #"
  "#. ###.# #"
  "#     #  #"
  "## # $## #"
  "# $*$$ $ #"
  "##@      #"
  "##########")

(add-solution
  "UUUUURRurrDLulullDRRRdrUllldRlullDRdddddrrULrdrruLLdllUUUUUluRdddrrrDu"
  "llllUruuurrddLUrrrdLLullDDldRuuulDrdddddrruLdlUUUUUluRRRRllldddddrrdrr"
  "uLuulLrrdddllluuUUUluRRRdLulDDDrrrdddllluUUUUluRRdlddrrrddLLdlUUUUUluR"
  "dddrrrddrdrruLLLLUruLLrdddlluuUUU")

(defmaze
  "########"
  "#....*@#"
  "#$# ***#"
  "# $ #  #"
  "#  $  *#"
  "#  $$ .#"
  "#     ##"
  "########")

(add-solution
  "LDDrDuluurDldDlLuuuRldddrruuLulllDDDRddrrrUUlLrrddlULuuUrrDDLruulldDrd"
  "dlllUUUUdRldddrrruulUlldRlddrURlluurrDrruuluLDrrruLLdlDllddrdrrrUUUddd"
  "lllluuurrururrdLddddlUdlluRRdrUUUUruLdddrUdldlllluurrDRdrUUUddllulldRR"
  "RRllldRRdrUUU")

(defmaze
  "#########"
  "#@*.*.  #"
  "#*...* ##"
  "# $#.#  #"
  "#   #   #"
  "#   $$  #"
  "## $$ $ #"
  "#  #    #"
  "#########")

(add-solution
  "DrrrRurDllluRdllluRdrrrrDDlDLLullUdrrdrrurrdLrddlUUddllURdrruuuulullll"
  "DlddRdRRlluluuruulDDrurruLrdrrdrddddlllulluUUluRddrDllUruuRurDrruLrddr"
  "ddlUUrddddlUdlluRdrruuullDLddrUdrruLUddlluRluLulldRdRRuLrddrrruulLrruu"
  "lDrdddllluluuluurrrurDllluRdllluRRdllDrdrddlUUUUluRdddrddrdrrruuulldLL"
  "dlUUUUrrrrDrddLLLLdlUUUluRRRllddrdrddrrUruLLLLdlUUUluRRlddrdrddrUruLLL"
  "dlUUUluRddrdrrrruulDrdLLLLdlUUUddlUU")

(defmaze
  "##########"
  "# #@$ ...#"
  "# $ $ ** #"
  "#  $# .*.#"
  "#   $ #*.#"
  "# $$ #.**#"
  "#  #     #"
  "##########")

(add-solution
  "RRdddLddrrrrULrUULDrddlllluuruRluuRDlddlddrrrruuuULDrdddlllluuruuRlddl"
  "ddrrrruuulLrrdddlllluuLUlldRlddrURlluurDurDrrUruulllDDuurrrddlUruLdddl"
  "LulldRlddrUURuuRRddLrurrrdddllllULrdrrrruuullluurDllldDlluuRRRRddLddrr"
  "rruuulLrrdddlllluuLuurrurDlllddrrURuLddlddrrrruuuuLrddddlllluururRurDD"
  "ullluurDldRRurDllldlluuRRurDldRRllullllddRUluRRRRurDllldldlddrUUUluRRR"
  "RRldRldldLUUluRRRurDlddllllddrUUUluRRRRlluRR")

(defmaze
  "###########"
  "#         #"
  "# $ $ $#  #"
  "#####  # ##"
  "##  ##$ $@#"
  "#   ..$  ##"
  "## #..#   #"
  "#  #..#  ##"
  "#  $*.#   #"
  "#  #. $ # #"
  "###########")

(add-solution
  "LdlLLddddlUUruuLLulDDDlddrURRuruullldDldRuuurrrddlUruLddddrUUluuLulDDD"
  "lddrURRuuulldDldRuuurrrrrruuuulllddrDuluurrrddddlLLLddrUdlddrUUluuLulD"
  "DDlddrURRuuulldDldRuuurrRRuuluurrrdddLruuulllddrDrDDuLLLddrUluRRurrddl"
  "UllddlddrUUUluLulDDDlddrURRdrUUluulldDldRRluuurrRRuuluurrrddddLLLLddrU"
  "luRddddlUUruulLulDDDldRRluuurrrddddlUruuuRuuluurrrddddLLLLLulDDDldRRRd"
  "rUUdllluuurrrrrruLruuulllddrDrdLLLLulDDDldRRRlluuurrrrrrdddldLLUdrruUU"
  "ruLLLLLulDDDldRRRlluuurrrruuluurDDDrdLLLLulDDDldRRluuurrrruuluulldRRur"
  "DDDrdLLruulullulldRRRRurDDDrdL")

(defmaze
  "###########"
  "## #  ##  #"
  "#    #.*. #"
  "# $  .*.* #"
  "## $ $# .$#"
  "#  $   #..#"
  "### $## *##"
  "# #   $@* #"
  "#    #    #"
  "## #  # # #"
  "###########")

(add-solution
  "LrdrruLUddlluRurUrUULuLrdrddlUlULLLddrUluRddlDlddrUUUruulDLDDrUluRuulD"
  "ulldRRDrddlUlUdrrdddlUUrurUlLuurDDluulldRRdrDDlUlUdrruLuurDDlddddrURRd"
  "rrULrUUruuLuurDDDulDDDlddrUUUUUddddlLLLdlUUUUrrdLDldRRRRdrUUUUdddllllu"
  "uulDDruuuulDulldRRDrddlUlUdrrdddlUUruuLuurDDDDldRRRRdrUUUddlllluuuluul"
  "ldRRurDDDDldRRRRdrUUdllllluUlURurDDDldRRRRdrUllllluuuuulldRRurDDDDldRR"
  "RR")

(defmaze
  "###########"
  "#@$...#  ##"
  "##****.   #"
  "##. .###  #"
  "#       $ #"
  "#####$#$  #"
  "# $     $ #"
  "#     # # #"
  "###########")

(add-solution
  "RRRDRRRurDrddLruulDDuullllullDRddlUruRRRRurDrddlLLLLUdrrrrrddLLLLUdrrr"
  "ruuuullllLddRddrrrruulUrdddlUUddllluuRRDulllulldRRuuulDrddRddrRurruuuL"
  "LLLrrrrdddlldlluuluUddlluRuulDrddrrddrruuLLrrdrruLUddldlluurRdrruuuLLL"
  "LLLddRRddrrurUUruLLLLLrrrrdddldlluulluUddrrRRddlldlluRdrUUdrrurrdLLrru"
  "uulDuulllllddrRRRddLLdlluRdrUUdrrrruuululllllddrRddrrrruLUUdddllluuRRl"
  "lddrrrruuulDllllluurrrrurDrddddlllldlllluRRRdrUUdrrrruuuullllllddrRddr"
  "rrruulUruLLLLLrrrrrddddlUUUruLLLLrrrddddllluuRRllddrrUrUUruLLLrrddddll"
  "luurRdrUUruLL")

(defmaze
  "#########"
  "#..... @#"
  "#**# $$##"
  "#.  ##  #"
  "#*.$    #"
  "# $ $#  #"
  "#  $$   #"
  "# #     #"
  "#########")

(add-solution
  "lDDDDrddllllURdrrruulullLDurrrrdddllluUluRRllullDRRdrddlUlURuulDrddrdr"
  "rruuuuluullllDDrdddrdrrruulUrdddlULrruulDrddlllluuuulldRuuulDrurrrrdDD"
  "rdddllllulUUUUluRdddlUrrDlddrdrrrruuullLLulDrrrruuulldRurDDrdLLLLulldR"
  "uuulDruRRdrrdrddddllllulUUUUluRRldddddrdrrrruuulllLrrrrdddlllluluUUUlu"
  "RdddlUUrddddlUUUrdddrdrruLULrddlUlUUUUddddrrrdrruuullLLrdLLdlUUdrUUrdr"
  "rrrddLLLdlUrrrUruLLLLrdLLdlU")

(defmaze
  "##########"
  "#    # @ #"
  "# $$ # $##"
  "# $ $ $  #"
  "#   ## . #"
  "# $#.**. #"
  "## .... ##"
  "##########")

(add-solution
  "DDDlUdDLdlllUluuRDluuurDRllddrUdrURRRddldlllUUrurrrdrrdLdLLuRdrUluulll"
  "dlddrRRurrruuLLruulDDrrddlUUddlldllluururRRDullldlddrrruRurruLLuurDDld"
  "dldllluururRlldlddrrruruUrrddLLrdLLLrruuuruulDDllldlluRRRRRDDrdLLrruru"
  "uLulDllllluurDldRRRRRDDrdLruruuLulDlllluurDldRRRRDDrruuLDuulDDullluurD"
  "ldRRRurD")

(defmaze
  "##########"
  "#        #"
  "# # $$##.#"
  "# $ $@##.#"
  "#   $$ #.#"
  "## $$ $..#"
  "#   #....#"
  "##########")

(add-solution
  "DLLuuurrrrrdddddlllURuLrddrruuuuulllDDDuuulldddRluuurrddLdlUrruurrrddd"
  "ddlluuLrddlUUUlldRurDrddrruuuuullllllldddRRlddrUluluuurrrrrrrdddddlllU"
  "LULrdrruLrddrruuuuulllDDDuuullDDRluurrrrrdddddlluulUdrddlUlLUUrDldRRdr"
  "rruuuuullllDDrDrDRdrUUUddllulullddrRRRdrUUdllluulDldRRRRdrUllldRRllulu"
  "uuurDDDldRurDldRluluuuullldddRRlluRRRurDDDuulDlddlUrRRurDlluuuulDDDull"
  "dRdRRRRRllluulDldRRRR")

(defmaze
  "##########"
  "#@$   ...#"
  "##$#.*.* #"
  "#  $. #.##"
  "#   #    #"
  "#  ## #  #"
  "# $ $ ## #"
  "## $$#   #"
  "#      # #"
  "##########")

(add-solution
  "RDDlddrDluuurDuRRuurrdLulDrrurrdLddllULuurDldLdlluRuuRRddLdlUdlddrUUru"
  "rurRurrdLddrdddlldllllUUluurrururRurDllldldlluRRdllddrUUddddrrrrurruul"
  "ullUdddLLrruurrrdddlldlllluUluurruRuuRRdrDulullddldllddrddrrrrurruuuLu"
  "uruLLrdddrdddlldlllluuluurruruRRurDllldldllddrrrruuUlldlluRRRuullDurrd"
  "RRuLLddldllddrrrruuUluurrdLrurrdLullddlldlluRRRlluuRRRRRllllldddlddrdd"
  "rrrrurruuluUdrdddlldlllluurrruuUUdllluurrRRldddRlddlllUUluRRRldlddddrr"
  "UdrrurruulUllUUddddLdLdlUUUUluRRRldldddrruruuUdddlLdlUUUluRRluurrR")

(defmaze
  "##########"
  "#    # .##"
  "# $  #...#"
  "###$  *.*#"
  "# $ ## . #"
  "#   ### ##"
  "# $$   $ #"
  "##     #@#"
  "##########")

(add-solution
  "uLuulUllululldRRurDlDRRRdrddLdlluRdlllURuUUruLddddlluuRlddrruUdddrruLd"
  "lluluurrUruulDDDurrrdrrULLLLrrruulDrddddLdlllluluurrUrrRllldllddrUddrr"
  "uLrdrruruuUUdlllldDuuruulDulldRRurDlDRRRdrddldlluRdlllURuUUruLddddlluu"
  "RlddrruUdddrruLdlluluurrUrrrdrddLruuluurDDlulluulDulldRRurDlDDllddrdrr"
  "uLUUURRRdrUllluulldRurDlddllddRdrUUUURRRdrruLdlUlluulldRurDlddddrdrrur"
  "UdlLLdlUUUURRRRllluulldRurDlddlldRdrUUURRRlluulldRurDldRR")

(defmaze
  "###########"
  "## # #   ##"
  "#    #  $ #"
  "# # $ #   #"
  "#.* . #$  #"
  "#****$    #"
  "#@*..#  # #"
  "###########")

(add-solution
  "UURRRDRRRUrruuLulldRurDrddllUddlllulllddRURuuullDDRdrUdRRRdrUUrrdLLLLL"
  "ullluurrrDrDululllddrddlUUrrRdLullddrUluRdrrRRRurrdLLLLLuurDldLLuRdrRR"
  "RurrdLLLLLLulDrrrrrrruuulDDrdLLLLLLrruululllDurrDrrddrrruuuulldRDDrdLL"
  "LLULuurDrddrruuuurDDDrdLLLL")

(defmaze
  "###########"
  "#      ####"
  "#    ##  ##"
  "#  ## $ $ #"
  "##     $  #"
  "##$.##$#$ #"
  "##.*#    ##"
  "#..$  $####"
  "#.....  $@#"
  "###########")

(add-solution
  "LLLLulLUlUUUluRddddrddlUrrrdLLulUUUrDDldRluuuUluurDDDDrdddrruurrrUruuL"
  "DLLLLLddddrruurUdlddllulldRRuuulDuUUluurDDDDrurrRddrrUruulDLLddrrUruul"
  "ulDrrddldlluuLuRdddlddrUUUUlLLdddRldlluRRuulDrddlluRdrUUluUUluurDDDDDr"
  "ddlluRdrUluuuuuurrdLulDDDDDDrdLuuuurrrrddddlLrruulDrdLruuuulllldddrRll"
  "uuurrrrddrruruulDLLLLLrrrddlDllluuUUluurDDDDDuurrrrrruulDrdLLLLLrrrddd"
  "lllLuuUUluurDDDDDDrrrrurrUruLLLLLLrrrdddlllluuUUluRddrrruRdrruulDrdLLL"
  "LLrrrdddlllluuUUluurDDDDDuuuuurrdLulDDDD")

(defmaze
  "########"
  "##     #"
  "# #  $ #"
  "# $   ##"
  "#.*#$ ##"
  "#...$$##"
  "# .. $@#"
  "########")

(add-solution
  "LULLdlluuuRRururrdLullddlldddrrurrUUddlldlluuurrRDrddLLuRuullldddRluuu"
  "rrrddlLdlUrrruullDurrddlLuuruurDurrdLLDllddrrUrULdddrUUluuurrdLulDDDrU"
  "dddlUlluuRRuulDrdDrddlUlluuRururrdLullddldddRurruUddllluurRDrdLLdlUrrr"
  "dLLruuuluurDDDrdLLruuuurrdLulDDDrdLruUluurrdLulDDDDrdL")

(defmaze
  "###########"
  "######   ##"
  "##  ##$  ##"
  "##  # $$ ##"
  "#       ###"
  "#. #$ # ###"
  "#.#. #  $ #"
  "#...# $$#@#"
  "# ..    $ #"
  "#  ########"
  "###########")

(add-solution
  "dLLLLLLuurUruLLrddldlluurUdlddddrUrrrrrUUlDrdLLLLuurururrDDuuUruulDDDl"
  "luRdrUruullDDlddldlddrrruurUUddlddllluuruulldldddRRuururuuruurrddLdLLr"
  "ruruulldDrruLddddlddrUUUUUllddldlddRRurrdrruuLrddllulldlluururuurrddDu"
  "uulldRldldlddlluRdrrrururrrddllLuurDldLLLUllddrUluRluururRRurrDDDrrddl"
  "LLLLLrrruurDuuuulldllldlddddrUluuururuulDrdrrRurDDDlddllllluuuuRRRRurr"
  "dddrrddlLLLLLLdlUUUddrrrrruurDurrddlLLLLLLdlUUdrrrrrruuuuulldRurDDDDur"
  "rddlLLLLLLdlUrrrrrruuuuuruulDDDDDDurrddlLLLLLrrrruuuuuuulDDldRurDDDDur"
  "rddlLLLL")

(defmaze
  "###########"
  "##  #@....#"
  "##   $$.**#"
  "## $  $$#.#"
  "#  $# #  .#"
  "#  $ # ## #"
  "#         #"
  "###########")

(add-solution
  "rrrrDDDllULUdrdrruuullDlLLdRluLulDDDlddrUUddrruLdlluurURDuurrdLulDrrur"
  "rurrdLLLLdllldlddrrrrrrrrUUllUdrrddlllllllluururrrurRurrdDuulldRurDlll"
  "ldllldlddrUdrruLUddlluRluRuRuRRRRRlllllulDDDlddrUdrruLUUluRRRRRllllddd"
  "dlluuRlddrruUUluRRRdLruuRRRlddLUllllddrUluRRRuRRddlUdLLullddrUluRRRRdr"
  "UdlllldlldRdrUUUluRRRR")

(defmaze
  "#########"
  "#..$.#  #"
  "#....** #"
  "# $.##  #"
  "## $  $ #"
  "####$ $ #"
  "# $   $ #"
  "#     #@#"
  "#########")

(add-solution
  "uuLruulDruuulDLLuLLdldRuurrdLLrrrrrddlUrdddLUUruuLLLulldddRRRllluuurrd"
  "rrrddddlLdlluRURuLdddrUrruuLUdDrdLLdlluRuuRlddRUrruulDrddLLuluRddrruLL"
  "rruulDruuulDllulldRlddRUUddRRddrruuuullLrrrddddlluullluuuRDlddrrrdddll"
  "uRdrUrruuuulllllddrrDulluurrrrrddLUruLLLLrrrrddddlUUddldlluRuRlulluUrr"
  "rrrddlUruLLLLrrrrddddlUUUruLLLrrdddlluRdrUUruLLrdddllddrUluRddlllluRRR"
  "drUrUUUruLddlldRdrUUU")

(defmaze
  "###########"
  "#  #      #"
  "# $# $ $$ #"
  "#  ##$#$  #"
  "#  $    #@#"
  "# $## #  ##"
  "##  # .## #"
  "##$ #.*. ##"
  "# ..*. *. #"
  "# # .. #  #"
  "###########")

(add-solution
  "uuulDulldRlulldRDDDDDRdRldlUdlluRRRuulDrddlluRdrUUluuuuuurrdLulDDDDDDr"
  "dLuuuuuurrurrdLLLulDDDrrUdlluurRddllDDDrddlUUUUUUdrruululldRdddddddllu"
  "uulUluurDRRRuuurrdddLruuullddDDDDrddlUdlluRRlluulUluRRRRddDuuuuuurrddd"
  "LruuullddDlllddrddrdrrULLdlUlUUUluRRRRDDDuuulllddddrrdrruLuuuuuuurrdLu"
  "lDDDDDDDrdLuuuuuuurrddLruulldDDDDDuuuurruuurrddLruLLLulDDDDDDuuurrUruL"
  "LulDDDDDuulllluuurDDDDDDurD")

(defmaze
  "############"
  "#@##. #  # #"
  "#**..$#    #"
  "#. $..#  # #"
  "# $#.$#$$$ #"
  "#         ##"
  "#  ##  #   #"
  "############")

(add-solution
  "DRdDldRuuulDDrdRRuuLrddlluluurDrrddlldlUUdrrrRRRRUUlDuuurDrrddLLUdrruu"
  "lLddddrUlLLLLuurDluuurDDllluRdrruulDrdLddRUUddRRRuulDruurrddLLddrUlLLd"
  "lUluuulldDlddrUUdRRRRRRuulDruurrddLLddrUlLLLLuurDldRluuulllDDrURurDlll"
  "uRddddlUUrurrrddlLrruullldlddrUrrruulllulDDurrrrddlUruLuurDDlddllUluur"
  "RllddrdrrrRRRuulDruurrddLLddrUlLLLLuurDldRluuUlllddrURurDllluRRdldddlU"
  "UrurrrddlLrruullldlddrUrrruulllulDrrrrddlllUUluRdddlUrdrrrRRRuulDrdLLL"
  "LLrruullldlddrUUUrrrddrrruuuulDDDrdLLLLLrruullldlddrUUlURRlddrrrrrruuu"
  "rrddLLulDrdLLdlUUdrrrdrUluuurrddLLulDrdLLLLLLdlUU")

(defmaze
  "##########"
  "#     ####"
  "#   $$####"
  "#.*  .####"
  "#$*$#*####"
  "#@*... .##"
  "#####$ .##"
  "#### $ $ #"
  "####     #"
  "##########")

(add-solution
  "RRRRRdddrUdllluRRuulDullUUrrDDllllUURuurrDLDllddrrrrrdddlluRUruLuuLLrr"
  "dddddrUdrruLLdlUUUrDldRluuLLUUrrDDrddlUruLuuluurDDDDDuuuluulDDDuuullDR"
  "RRurDDDDrdddlluRUUUddddrUdrruLLdlUUUrDlddrruLdlUU")

(defmaze
  "############"
  "#@#  #     #"
  "#  # #$#$# #"
  "# *$       #"
  "#..##$#$   #"
  "#..#. #  $ #"
  "#....# $$$##"
  "#          #"
  "############")

(add-solution
  "drDDDuuRRRRRRdrruuullDDlDRlulllllddlddrrrrrrrUUddlluRdlllllluuruurrrrr"
  "dDuuruurrddddLLUUllllllddldRdrrrrrrUUrruuLLLLLLLrrrrrrrddllddlllluurUd"
  "lddlluRdrUllluuuurDDDldRRdrrrrruuUlulLLLrrrrdrdddllllllluuuuurDDDldRdr"
  "rrrrruurruulLLLLLLrrrrrddddllllllluuuuurDDDlddrrrrruRUUruLLLLLrrrrdddd"
  "lllllluuuuurDDurrrrrrddrUruLLLLLLLrrrrddddlllllluuuuurDrrrrrddddrUUUru"
  "LLLLLLLulDDDuurrrrrrrrruullDurrddlLLLLLLLulDDurrrrrrruullDurrddlLLLLLu"
  "lDrrrrrrrddddrUUUruLLLLLLL")

(defmaze
  "########"
  "#      #"
  "# $   $#"
  "##$ $  #"
  "#@$*#$ #"
  "# #..$*#"
  "# .....#"
  "########")

(add-solution
  "ddrrrrrULUUrDlddlluUURuurrDLDlldddrruUrULuullDDDuuurrddLruulldRlulldRu"
  "rrddrdddlllluuRURRuurrDLDDrUlllldlddrrrrrUdllllluururrrdDrdLLLrruuuuul"
  "lDRurDDDDrdLLruuulululldRRRurDDDDrdLuuululldRRurDDDuurDDluuuurDDD")

(defmaze
  "########"
  "#..  $ #"
  "#. $$$ #"
  "#.$# $ #"
  "#*.. # #"
  "#.$ $$ #"
  "#..#@  #"
  "########")

(add-solution
  "rruuuuuLLDLLDlDDRluuruulDDrurrurrdLLLLdDrrUdlluurRurrdLLddlluuuRldddrr"
  "uuLulDDlDRuuulDDrurrrrddddllUUUdddrruuuLuLLLulDrrrrrddddlluuUruLLLddDR"
  "luuurrrrddddllUUUruLLrruLLLDDurrrrdddLdlUUUUruLL")

(defmaze
  "##########"
  "#    #  .#"
  "#  $$$ #*#"
  "#   $ ...#"
  "##$ $***.#"
  "#@$    ..#"
  "##########")

(add-solution
  "RUUluurrDDRRRuurrDDLLLLLuullddRddRRULrdRRULuRuurrddDLruuullddRdrUUdlll"
  "ddlluURRRRdrUllllllluurDldRRRRRRlldRRldllluRRRllluluurrDulDDRRRuLdddll"
  "lURRRlluulldRRRRlluurDDldR")

(defmaze
  "###########"
  "#  ##  #@ #"
  "#.* #  ## #"
  "# * #   $ #"
  "#.* ## $###"
  "#.$ # $  ##"
  "#.*   $   #"
  "# .  ##   #"
  "###########")

(add-solution
  "rddLLDDrddlUUdLLLdllluuRDldRuRRRRRdrruLLLLLLLuluuRDrDulldRddlUUrdRRRRR"
  "drruLLLLLLLuurDluuluurDrDDldlUUdrdddlUUrdRRRRRdrruLLLLLLUluUrDDllddrUU"
  "UddRRRRRdrruLLLLLLullddrUUrdrruRuUluurDDDDldRRdrruLLruLdLLLLLdlUrrrrrU"
  "UUluurDDDDrdLLLLLrrrruuurDDrdLLLLLrdL")

(defmaze
  "###########"
  "#  #   .#.#"
  "#     $* .#"
  "# $ $ *...#"
  "# $ $$$ #*#"
  "#####@#  .#"
  "###########")

(add-solution
  "UURRllUluRRDRdddrrUULrULLDuLulldRdRRuLulldldRRllulldRRururrdLLrrrrrddd"
  "llUUddrruuLuLDuLulldRdRRuLulldLdRuurrdLLrrrrrdddllUUddrruuulLLullddRdR"
  "lulldRRluuurrdrrrdddllUUddrruuulLdLLdRlllulldRRRRuLuurDurDrrrdddllUUdd"
  "rruuulLLulldddrRUrurrdddllUURurDDulldllluRRRRurDlldllllluRRRRRRlllllll"
  "uurDldRRRRRRllluurDldRRluurDRRllldllluRRRR")

(defmaze
  "############"
  "#          #"
  "# $$$$@$$* #"
  "# $  ## .* #"
  "# $$   ##. #"
  "#   #*...* #"
  "# ##...*.# #"
  "############")

(add-solution
  "ullDDuulllddRRlluurDurrdLrurrdLrurrDLrurrdLDuulldRlulldRRllulldRRllull"
  "ddddrrUUddlluuuurrrrrrrrrddddLLdLLuulLUdrrddrrurruuuulllllllllddRRlldd"
  "rUluuurrrrrrrrrddddlldlluullLrrrddLUruLrddrruLdlluurDldRRurrruulDrdLLL"
  "dlluurDldRurrrruuuulDDDrdLLLLulDrrrruuuulldRurDDDuullulldRRRurDDullllu"
  "llDRRRRRlllllulldRRRRRRlllllldddrruUddlluuRuRRRurrrrrddddLLLLulLUdrrdr"
  "rrruulDrdLLLrrruuuulDDuulldRlulldRRllulldRRlddrrDrrrruulDrdLLLrrruuuul"
  "DDDrdLLruuuulldRurDDDrdLuuullulldRRRurDDDuullllllldlddrUUluRRRRRRRurDD"
  "ullllllldRlddrUUluRRRRRRurDllllllddrUluRRRRR")

(defmaze
  "#########"
  "#     *@#"
  "# $$$**##"
  "#    #. #"
  "##$# ...#"
  "#  $ #. #"
  "# $ ##* #"
  "#   ..$ #"
  "#########")

(add-solution
  "LDDrddddLLrruulDrdLruuuulDDuuuuLLDRurDDrddlUllUllluurDldRuuRRDrruLLrrd"
  "DrddlUlldLddlluuRUUluurDDuRRuLrrrdDrddlUllULuurrrdDuulllddrdrRllululld"
  "RururDurrdLLrrddlldLddlluRluRUUluuRRRdrrddlldLrurruullulllddrdDuuluurD"
  "DuurrdLrrrddllULuRlullddRluRurrdddrruuuLLrrddddrddlUUUlluuurrdDrDLddrU"
  "luuuulldRurDDDDuuulldddllUUluurDRRRurDDDuulldddlluUluRRRRurDDulllldRld"
  "drruUluRRurDlldddlllddrrUdlUUUUluRRRRldddLdlUUUluRRRlluRRR")

(defmaze
  "#########"
  "##   ####"
  "#    $  #"
  "#   ##  #"
  "# $    $#"
  "# $# *..#"
  "#  #$**##"
  "# ##@...#"
  "#########")

(add-solution
  "rrUUrULuuLLLrrrrDlddddllURdrUUULDDldRRuuuUruLLLullDRddRDrruuulLulDrrrd"
  "ddlluRdrUUruLLLulldlddRRRRlllluururrdrrdDDDldlUruulllluurDldRRRRllluuu"
  "rrdrrdDDrUluullLDldRRRlluurrrdDrdLulDDlUrruuulllddRRlluurrrdDrdLulDlul"
  "llddrUluRRRRlluurrrdDrdLullluuulDDldRRRRlluurrrrDDulD")

(defmaze
  "#########"
  "# ..#   #"
  "# ..*. ##"
  "##$  $  #"
  "#   ##  #"
  "##### $##"
  "#   $ $ #"
  "#@      #"
  "#########")

(add-solution
  "rrrruLLrrdrruLLrUUULuurDDluLLulldRRddlUruRRdLLuulldRRddlUrrrrDDlddrUUU"
  "ruLLuurDDlLLuRlulldRRddlUruulldRdrrrrDDlddrUUUruLLLLdlUUluRddrrrrddddl"
  "lluRRdrUUUruLLLLdlUUdrrrrddddllllluRRRRdrUUUruLLLLdlU")

(defmaze
  "#########"
  "#.##   ##"
  "#.*.$ $##"
  "#.*..   #"
  "#*#*. $##"
  "#@$  $ ##"
  "##$# $  #"
  "#       #"
  "#########")

(add-solution
  "UURRuRurDDrDDuululldllURRRurrDLdLddRUluLulldddRRUrrrULdlldlluuurrRDrdd"
  "LddrUdrruLLUUrDllLuRuuurrDDDLdlUrruuulldllldddRRurrdddlllUdrrruuulldRu"
  "rruuulldllldddRRuULulDrrdrrruuulldLLrrurrdLLLddrrULLLrddlddrrruUUruLLL"
  "ddRllddrrUUdrdrruLLUUruLLrdLrrDlddrruLdllUUddrUUUruLruLL")

(defmaze
  "##########"
  "#@*..  # #"
  "#*.$. #  #"
  "#..#.$   #"
  "# $ #  $ #"
  "#$ $  $  #"
  "#   #  # #"
  "##########")

(add-solution
  "RdRRdRRdlddrUUluluurDDDrdrruLruulDLrrddllUdrruulDllURlluurDldRdrddlUlL"
  "dllURRRRUUlulldDrdrrurrruulDLLdddrUllluluurrurDDrrrddLLLrrruulllDurrrd"
  "dllUdrruulDlluRdddlUruululllddrdRRdrUUdllluluuuRdrrdDuulllddrddlUUUlDr"
  "rdrRuuullullDRddrdrruuuuLLrrdddrrruulDLdlddrUlllluuuRurrdDrrrddLLLLLul"
  "uuluRRdlddrddlUUUUluRdddlUrdrdrrrrruulDrdLLLLLdlUUUUddrdrrrruulDrdLLLL"
  "dlUUUdrdrrruulDrdLLLdlUUddlUU")

(defmaze
  "###########"
  "#     #@ ##"
  "#   $###$ #"
  "##  $  $  #"
  "#     #.$ #"
  "# #$ #.** #"
  "# # $#*.*.#"
  "#  # ... ##"
  "###########")

(add-solution
  "rDrddddLdllllUlUUUdrruLdlluuurrDDLruulldRddlUrdddrdrrrruruuLDruuulDLLL"
  "LLuurDlulldRRddlUrRRRRRurDDDlUruLLLLLLdddrdrrrrUUruulllldLulDrrurrrrdd"
  "lddlllluUUrurrrrddlLDrdLLrruuruulllldlluRRRRDDDrdLruuruLruulDllllluurD"
  "ldRRRRDDDruruLruulDllllluulldRRurDldRRRRDDrruLruulDDDuullldlddlUUlURRR"
  "RRDrruulDDlulllluulldRurDldRRRRRdrruulDrdL")

(defmaze
  "##########"
  "#    #   #"
  "#.#$$##  #"
  "#.$  $ $ #"
  "#.$      #"
  "#..*##$$ #"
  "# ...*@# #"
  "##########")

(add-solution
  "LLLUlluuuurrrDDRdrrUdlluluulllddRRllddrrURULddlluuuurrDDLDlUrrrrRdrruL"
  "LLLdRRlluLLdRddrrrUdllluuurrdLrurrrddLULulldRRllulldRRlddrrrUrruuLLLLd"
  "lllddRRRluururrrrddllULLLLuRdrrrdrruLLrruuulDDrddlluLuRdlLLuRRdrdrruLL"
  "LLulDDLdRuuuuurDDrdLulDDuulDDrurrrrruulDrdLLLLLrrrruLLLulDDulD")

(defmaze
  "###########"
  "#.........#"
  "# #$####. #"
  "# $@# # $ #"
  "# $$$     #"
  "#$$  ## $$#"
  "#    #    #"
  "###########")

(add-solution
  "DRRRRddrrUULUUruLdddrUllDlllddlllUURRRRRRuruuLLLLLDDuurrrrrddldlllddll"
  "luuURdlddrrruuLUlldRlddrURlluurDurDrrrrurrdLuuurDlddrddlUUUUruLLLLrrrd"
  "ddLLLLLulldRRRRRRurDrddllUdrUUUUruLLLrrdddLLLLLuUdllUddRRRRRRdrUUUruLL"
  "rdddlllllllddrUluRRRRRRdrUUUruLdddllllllddrUluRRRRRdrUUUddlllllddrUluR"
  "RRRdrUUdrUU")

(defmaze
  "#########"
  "#     #.#"
  "# $$$ #.#"
  "# $   #.#"
  "#   $ #.#"
  "#  $ #@.#"
  "#  $$**.#"
  "##   .*.#"
  "#########")

(add-solution
  "rddLLLULUlluuRluurDDuurrrddlLrruulDLullddddrrUUddlluuuurrrrdddLUruulll"
  "lddddrrdrUddrrruulDLdlUrrrdLulllulldRRRRRdrUUUUUddddlllllluuuuurrrrddl"
  "lDDldRRRRdrUUUUdddlllluuurrdLDldRRRdrUUUddllluuulDDldRRRRdrUUdlllllluu"
  "RRurDDldRRRdrUllldRRllulldRRRluuuruuulDDDDuuuulDDDDDRRRllllluuuuRDDDld"
  "RRRRluuuuulDDDDldRRlldRR")

(defmaze
  "#########"
  "#     #@#"
  "##$ $ # #"
  "# $  $$ #"
  "#  $  #.#"
  "#  # #..#"
  "#   $**.#"
  "##  ..*.#"
  "#########")

(add-solution
  "ddddddLUdLLUlldRurRdrruLLLLdlUrdRRluUULuuurrdLDDDDrdLrurrdLullllluuRRu"
  "rDllluRdrrrUluullDDRluurrdLddrUllldRRllddrrrrrruuuLLLLuurrDulldddlluRd"
  "lddrrrrrruuulLdLDuLUlldddrrRRRdrUUdlllllluuRDldRRRRRdrUllldRRllulldRRR"
  "luuuruLruulDDDDDldRuuuuuullDDRlDDldRRRRRlluuruulDDDDlluuRurDDullddrRRl"
  "lluuuuRurDDDDullddrR")

(defmaze
  "##########"
  "#.##     #"
  "#.* *#  ##"
  "#.**.. $ #"
  "# $..$  ##"
  "# # $ $$ #"
  "#$###    #"
  "#   @##  #"
  "##########")

(add-solution
  "lllUUURRllUURDRRRDrrULuullDLddRRlluurDuurrddLLrrrdDrddlUllUURuuullddRd"
  "rUdlddrUdrruLdlluuurrDLulDululLdRdRRuLdlUrrrrdLrDrddlUlUULuluurrrDDLdd"
  "dlUruuruullldLdddRRlUlulldRurururrrddldLdllUdrruLULLrrdrruLLLrrrruulDr"
  "dLLLrdddrUdrruLdlluuurrDLLddrUUruLLrdddrruLLdlUU")

(defmaze
  "##########"
  "# # #  # #"
  "#    $ # #"
  "#$$$   $ #"
  "#. *.##$ #"
  "#.** #   #"
  "# #. *.$##"
  "# ##...$@#"
  "##########")

(add-solution
  "LLuurruuLLLruulDllllDDuurrrDDLLUdrruurrdLullDlluRRRRdLLrrrrrddlUdlddrU"
  "UruuLLLulldRdDDRRurruullLrrrddlldlluuUlulldRdRluurDrruLLrrdrrrddlldllu"
  "UUlldRurrrrrddlldlllUrdrrurruullllDDldRuuulDrurrrrddlUdlddrUUruuLLLLdd"
  "dRluuurrrrddllDlluuuruurDldldlluRuRRdLulDrdddrrruruulLLrrrddldllluuUru"
  "urDldlluRddddrrruruulLLrrrddldllluuUlldRurrrrrddldllllUrdrrruruullllDD"
  "ldRRluuulldRurrruulDrdLLDDldRluuluRuRDrruulDrdLLulDDDuurrrrrddlUruLLLL"
  "ulDDluluRRRdrruulDrdLLulD")

(defmaze
  "##########"
  "#. .$.# @#"
  "#..*..$ ##"
  "#*$.##$# #"
  "#. #   $ #"
  "#* #$$   #"
  "#  $   $ #"
  "#     #  #"
  "##########")

(add-solution
  "ldLLuLLDlDDlDuruulDrururrdLrrDDDrruLrdddlUUlUllDRdRluurDrruLLdlddllluU"
  "UUluuRRRdrrDDuullulllddrddddrrruuRuuullullDDuurrdLullddrUluRdrrrrdddld"
  "dllluulUUUrdDuurrrrdddlluRdddlULrruulDrrUrrdLLLddllllUUUruulDrurrrrDDl"
  "dddlluLUUUluRRRllddddrdrruLLdlUUUUluRRlddddrrruurrrdddlULLLLdlUUUUluRd"
  "ddlUUrddddlUUUrddrrruurrDrdLLLLLLdlUUrdrrruulDrdLLLdlUrrrruurDrdLLLLdl"
  "UUUU")

(defmaze
  "############"
  "#  #@$  .  #"
  "#  ###$#.$ #"
  "#      $.# #"
  "## $$***.* #"
  "##   $.* . #"
  "#   ##.* # #"
  "############")

(add-solution
  "RRRRddLLLLLDlddrUluUrrrrDDLLrruullDRlulldRlddrUluurrrddLLURurrrrddLUdL"
  "LURdrruuuullDDDRdrrruuuuLLddDldRlulldRlulldRRurrruuurrdLrddLLLLuuuRRDD"
  "LdlUdLdlluRuRRdLrrRuuullDDuurrddLdLUllDlluRRllluurDldRdrddlUUrruRRdLdL"
  "ruulDllURRdrrrruuullDDuurrdddlLLLuRdrrruuulldDuurrdrruLdlddlLuRdrUllll"
  "dRRRlllulldRRRRllllddrUluRRRdLullddrUluRRlluluurDldRRRRRR")

(defmaze
  "##########"
  "# $..   .#"
  "#@$**.**.#"
  "## #. $ ##"
  "#  $$$   #"
  "#    ### #"
  "#  ##    #"
  "##########")

(add-solution
  "uRDuRRRRRllddlULulldRRRdrRULuLLdRdrruRuLLrdddrUllluRlllDDlddrUrrURuull"
  "ldDldRuuurrrddlLrrurrdLuluurrDLulDllldDrrUdlluuruRRdLddlluuluRRldddldd"
  "rUUrrdLullddrUluRdrruuurrddLruulldRluLulldRRlDDlddrUrrUruuuLLdldDldRuu"
  "ururrdddlLrruuulldRlldDrrUdlluururrDrruLLdLddrrUdrUdllUdllllddrUluRRRd"
  "LullddrUluRRdrUUrrdLLdlluRdrUlluuluR")

(defmaze
  "############"
  "# #  .*.$. #"
  "# $ *.*.*..#"
  "#   $ . $$##"
  "#   #### $ #"
  "##### $  $ #"
  "#@    #    #"
  "############")

(add-solution
  "rrrruRRdrrruuLULLuuLLddLLuuRlddrruRurRDLdLUrrRddrddlUUddlluRdrruulULDu"
  "ulldRlulLdRRllldlluRRRuulDrdLrurrdLrurrDLullllddlluuRRRRdLruRRdLulllll"
  "ddrUluRRRRRdrUlllllddrUluRRRRdLLullddrUluRRRdrrrrdrddlUdlluRUddrruuLuL"
  "LLLLullddrUluRRdrrrrrdrddlluUruLLLLLullddrUluRdrrrrrrdrdLdlUUruLL")

(defmaze
  "############"
  "#   #@#    #"
  "#.. $$$ $# #"
  "##.*# $$ # #"
  "# $.#      #"
  "##..#$  $  #"
  "# ... ## # #"
  "############")

(add-solution
  "DRDDrdRurruuullDDLdRuuulDllLLulldRRRRRRDrddlUlUdrruLuurDDluLLLLulldRRR"
  "RRRddllUdrruulLLLulldRdDrddlUUrddrrUUrruulDullLDDlddrrruuUdddllluuruur"
  "RdrrddLruuuurrrddddLLLuulluRlllddlddrrrUUUdddllluuruurRdrrddLrrrruuuul"
  "llDDrDuluurrrddddlLuullluRlllddlddrrrUUUdddllluuruurRdrrrddLLrruulllul"
  "lddlddrrrUdllluurDldRuuuuulldRurDDDDuuurrdrrrddrruuuulllDLLLLulldRurDD"
  "DuurrrrurrrddddlllllUUruLLLulldRurDDurrdddrUUddrruuLuLLLLulldRurDrrddr"
  "UruLLLLLDDDuuurrrrdddrUrdrruLLLLdlUUruLLLulDDDuurrrrrurDDLddrruLdlUUru"
  "LLLLLulDrrrddrUruLLLL")

(defmaze
  "###########"
  "##@#      #"
  "# $###$$# #"
  "#    $  $ #"
  "# #* *#   #"
  "# $...... #"
  "# # # #   #"
  "###########")

(add-solution
  "DldddRRRRRRurruuullDDDrruLdlUrrdddllUllluuRRRDrrddlULrruuuullDDDuuurrd"
  "dLDlUlllddllluuRRRRRRdrrddllULLLuurrrDrrddlULrruuuullDDDuuurrddLDlUrrd"
  "ddlUdlUrruulllllddrRRdrruuuuulllDurrrdddddllullluurRuurrrddddLLLLrrrru"
  "uuulllddRDuluurrrddddlLLruuluurDDDuuurrddddlLrruuLDruuulldDD")

(defmaze
  "##########"
  "#   #  #@#"
  "#  ##  # #"
  "#...*  # #"
  "# . # ## #"
  "# .##$ $ #"
  "##    $ ##"
  "###$$#  ##"
  "#        #"
  "##########")

(add-solution
  "ddddLdLLUUULLdlUrrrruulDrdLdddrruLdlUUULLdlddrRRurDluuuruulDDDDuulldll"
  "uRRRldlddrrRuuUlllluurDDDrurrruulDrdLLLdlUrrrdddlllUdrrrurrdLrddlUUruL"
  "dLUUUllllddrdrRRurrdddllllUdrrruUluuullllddrdRluluurDluuurDDrrrruulDrd"
  "LLLLrrrdddrruLdddllluuRRUUUruulDrdLLLrrddrrddlUruLdddllUluRRUUUruulDrd"
  "LLrddrrddlUruLdlUUUruulDrdL")

(defmaze
  "###########"
  "#   #@*. .#"
  "# $$#$$**.#"
  "#      *..#"
  "#   #$#. .#"
  "# ##  $ $ #"
  "#  #   # ##"
  "###########")

(add-solution
  "RDuRRDLrrddllURdrUllLLLLdlluRRRRRRRuulDullDRdLrRddLdlluRUUrrddLdlUrrur"
  "rdLuluUrDldRlullLLdlluRRRRRuulDrdLLLdlluRRRRurrDrddLLdlluRUURuulDrdLru"
  "rruLrddLrddlUUUlldRurDllLLdlluRRRRRurrdddlLdlUUdrrruuulldLrurrdLrddlUU"
  "ddlluuRlddrrruuuLDllLLdlluRRRRRRllddrrruUllllllluurDldRRRRRRllllluurDl"
  "dRRRRluR")

(defmaze
  "###########"
  "#     #.  #"
  "#  $ $ *# #"
  "# #  $*. ##"
  "#  $  $.  #"
  "# $$###***#"
  "# #@*... .#"
  "###########")

(add-solution
  "RRRRRlUULULLuLrddLUrrrdrruLUddlLuRlluurDRddlUlldRurrdrruLLuLullDRddRRu"
  "uLulllldddRRRllluuurrdDrdLuuurrdrddLrurrdLddllllUUUrRdRlulldddrrrruUll"
  "uuulllldddRluuurrrrddllDlluuurrDRlulldRRllddrrUrrdrrddllllUUlluururrDL"
  "ulldddrrRUdllluuurrdDrdLddrrrruulLulluullddddRluuuurrddrrdrrddllllUllu"
  "uuurrrrDrDLuullllddddrrdrrrruuruLdddrUdlllllulluuuurrrrdddRluurDrDDrdL"
  "LLrruuulldRurDDrdLLruulluuulldRlullddddrrURRRurDDrdLuulllldlluRRRRRurD"
  "DulllllluuRRDullddrRRRRurDurDllllluRRRllluurrDulDDRlldRRR")

(defmaze
  "##########"
  "#        #"
  "# ##  ## #"
  "## $$$  ##"
  "## $     #"
  "#     #. #"
  "#  ##$***#"
  "##$*...*@#"
  "#  . .* ##"
  "##########")

(add-solution
  "LUdLLdlllUUluRuRRdrDRdrdLLLrruulDrrrULdlluuluURuulDDrDrrDDrdLuuurDluuL"
  "dlUdDDRDrUUruLullluurDDrrdddllulUUdlluRRRuulDrdLddrUUdddrruuuLrdddlluu"
  "RurDDullluRluurDDldlluRdrUddrdrdrrULuurDlulLLrddRdrUUruLLLulllddRRUruL"
  "rdrruLrdddllUluRRurDDullddrdrrULuulldllluRRRRurDDDrdLuuullluRuulDDrdrr"
  "uLdldlluRRRurDDDuullluurDldRRurDDullllluRRurDldRRurDrD")

(defmaze
  "###########"
  "#.....  $ #"
  "# $$*# $$ #"
  "# .*. $ $ #"
  "###.#   $ #"
  "#  .* ##$ #"
  "##   #@   #"
  "###########")

(add-solution
  "rrruuuuuLLLLLDuLLDRurrrrrrddLUruLLLdDRluurrdLulLLLdDRUluRddRddLdlluRUU"
  "UlldRurrdrddLdlluRuUruuRRRddLLLuLDlluRRdrUdrddLdlluRUUluRdrrddLdlUUdrr"
  "uurrdLullLddrrUdlluuulldRurDDurRddldlluRuurrdrruLLLLulldRurDDurrrrurru"
  "LLLDrdLLLLulldRurDrrrrurrdLLLLLLrrrrrrdLrddlUUruLLLLLrrdRurrddlUruLLLL")

(defmaze
  "##########"
  "#  ..  # #"
  "# *.$$$# #"
  "# *..* # #"
  "# .$.. $ #"
  "# $$.$ $ #"
  "##  *# # #"
  "#  #   #@#"
  "##########")

(add-solution
  "uuuLLdddllUllUluuuurrrrrDDLULulllddddrdrrdrruuUdddlluURUrDlldlluluuuur"
  "rrdrrDullulllddddrdrruruUruuLLrrddlUruLddddldlluRURlddrUrrUrrdLLLLddrr"
  "UdlluurRuUluurDDDDlUdldlluRuRRdLulDrruUruulDlDurrdLddlluRdllluRRdrrrUr"
  "rdLLLLddrrUdlluurRuUluurDDDDllLulldRRRddrrUULULLrrdrddlluUruUrDDllluRd"
  "llluRRdrrrurrdLLLLddrrUdlluuLulldRdRluurDrrRuulDLLrrrdddllUUluRdddrruu"
  "LrddlluUlldRurrrurrdLLLruuuulD")

(defmaze
  "##########"
  "#     #@ #"
  "# $$   $$#"
  "# $  # * #"
  "# $ $ #..#"
  "# # #.*..#"
  "##   ..$.#"
  "##########")

(add-solution
  "rDLLrDDrddLUruUluurDDDDlUUdddLLuRdrUruulluLullllddRRdddRRlluuulluurrDu"
  "rrdLrrdrrddllluLULDrrdrrruullulullDRddlUruulllddRRlluurDurrdLddrrddllU"
  "UrrdrrruulDruuulDDluLulldRRRdrrddLdLLuulUdrddrruruullululldRddlUruulll"
  "ddRRlluurDurrdLddrrddrruruulluLulllldddRRddRRuuLrddlluUlluuurrrrdrdrrd"
  "dlLrruullululllldddrrrrDrrruulDrdLLrruuuulDDDrdLuulululldRRRdrruulDDDu"
  "lulllulldRRRRRdrruulDDlulllllddrUluRRRRRdrruulDlllllddrUluRRRRRurDDDuu"
  "lllllddrUluRRRRurDD")

(defmaze
  "###########"
  "# #  #@   #"
  "#   $ *$$*#"
  "# $   ..*.#"
  "#   ##$#$.#"
  "# $##  *..#"
  "#   ##  .##"
  "# #  #  ###"
  "###########")

(add-solution
  "rrrDDLLLLrrrruulllDldRluLLddlluRdrURurrdLLrruurrrddlDDrUUlLLuurrDDllDD"
  "RddlUUUUrrrddLLrruullldDuurruullDDulldRRuurrdLLrrdrddlUUddllUURuurrDLD"
  "llddrrruUluullDRdrrddllluUllLdlluRRRRRRuulDrdLruurrDLullddddrrruULDuuu"
  "llddRluurrdDllddrddlUUUUrrrddLruuluurDDDulllLLLdlluRRRRRRRurDllllluulD"
  "rdLdlluRRRRRRRDuuullDRurDllLdlluRRRRRllldRllldllddrUUluRRRRlllluRRRR")

(defmaze
  "##########"
  "#    ##  #"
  "# ### $  #"
  "#      $ #"
  "#  .# $$ #"
  "###*.* # #"
  "# # .*.. #"
  "#  #$**  #"
  "#@$   .* #"
  "##########")

(add-solution
  "RRRUURUrUdldlddRURUrruuLULLDRurrdddllUUluRddddlluRUruulDDDlddrUrUrruuL"
  "LDlDRllddrUluRuruulDDDllURdrUdlddrUluurrurrdddLLULrdrruuullulDDurrrddd"
  "lluUluRdddLUUlldRRuruulDrdLddrrruuuuLruulDrdddddllluuuuuRDlddddrrruuuu"
  "LLulDrrrddddllluurUUlLLDlluRRRRRddldllUdrruLrruulDrdLdddlUrrrruuuLLulD"
  "DDurDuurruLruulDrddllluuRDDDurruLLulDDDrurruuulDrdLLulDD")

(defmaze
  "#########"
  "#.. $ # #"
  "#.**..* #"
  "#.#.##  #"
  "#  . #  #"
  "##$  $$ #"
  "#  #  $ #"
  "# $$ $# #"
  "#     #@#"
  "#########")

(add-solution
  "uuLUUruuLLuLLDDDrdRldRlddrUUluuluuurrdrrddddLLrruuuullulldddrdDrrruuuu"
  "llLrrrddddllluullluuuRRDrrrrddddllddlllluurUdlddrrrruurruulDruuullllul"
  "ldddRddlddrrrruuluRdddlllluuruuluuurrdrrrrddddLruulDruuullllulldddrdRR"
  "RDlulldlddrrrUUddllluuruuluuurrdrrrrddddLLrruuuullllulldddrdrRuLrdDrrr"
  "uuuullllDDrddrrruulUrdddlUUddlluRluluuulldddRdRRlluluuurrdLulDrrdDrddr"
  "rruuuuLLLLLrrruLLLrrrdrrddlUruLLLLddrddrrUUUruLLLrrddddlluRdrUUUddllul"
  "uUddldRRRdrUUruuLLLuLDDurrrrddlUruLLLulDrrrdddlllldlddrrrrUUluRdrUUUru"
  "LLLrrddddlddlUUddllluururRRdrUUUruLLrddddlddlllURRdrUUluRdrUUUruLdddll"
  "llDldRRRdrUUluRdrUUU")

(defmaze
  "###########"
  "##..###.# #"
  "## .$ . # #"
  "## .*.. # #"
  "# #.#*##  #"
  "# $   ##$ #"
  "#  $#$$   #"
  "# $$    $ #"
  "##@###   ##"
  "###########")

(add-solution
  "URRRRdrrUdllulllUUUURlluurDRRRdrUllldRlulldRddddlluuRlddrrrrrdrruLLLLL"
  "uUrrDullUUluurDrrdDDlluUluRdddrruuLrurrdLLddlluUluurDRRdddlluUluRdddrr"
  "uuLrddllddrRUUddRdrruLuLrddlUllluuuUluurDDrruLrdrruLLdllDDrrUdlluurRur"
  "rdLLddlluuluurDRRdddlluUluRdddrruuLrddllllddRRUdRRRlUUlLuUluurDDrruLrd"
  "rruLLdllDDrrUdlluuluurDDrRuLdlUrrrrdLLddddrdrruLuLrddlUlUUUUruLLLdlUrd"
  "RRddllUdrrddrruLdlUUUUluRllldRddrrdrrdrruLLLdlUUUddrrrruulDrdLLLdlUUdd"
  "llUUUUddddlluRdrUUUddlluRdrU")

(defmaze
  "##########"
  "####..   #"
  "#.....$$ #"
  "####..#  #"
  "#   #*#$ #"
  "# ## $   #"
  "#  $@#$$ #"
  "## $$   ##"
  "#      ###"
  "##########")

(add-solution
  "uRRDulUUluurrrrddddLruulDruuulDulllddrURurrdLLLLLLrrrrrrddlUruLLLLLruR"
  "drrrddddLUUUruLLLLrdddRluuurrrddddldLUrUUUruLLLulDrdddRdddllURdrUUrUUU"
  "ruLLLdddRdddlllluRRRdrUUrUUUruLuLLddddlDurRdrUUUULulDDurrdddddldllluRR"
  "drUUdlluurRddlllluRdrrrruUUUruLLulDrrdddlllDldRRdrUUdlluurRdrUUUruLLrr"
  "uLL")

(defmaze
  "###########"
  "#..... # @#"
  "#..## $#$##"
  "#. # # $ ##"
  "#..#   $  #"
  "# $   $   #"
  "#$$$# $  ##"
  "#      #  #"
  "###########")

(add-solution
  "lDDDrdLdlLdlllllUUddrrULURRurrRDullldlddrrruUruLdddlllluuUUluurrrrrDDD"
  "DLLLrrruuulullllddrdDrrrrrrruLuLrddlllllluuluurrrrrdDulullllddrUluRRRl"
  "lddddrddlUUUUUluRRlddddrrrrrruulDrdLLLLLdlUUUUluRddddrrrrruulDrdLLLLdl"
  "UUUUdddlUUUrdddrruRdrruulDrdLLLLddrrrUdlllllUUUUdddrUUUddrrrurrdLLLLdl"
  "UrddrrUruLLLLdlUUdrrrrrrrdLLdlUruLLLLdlU")

(defmaze
  "###########"
  "#         #"
  "#..#   $$ #"
  "#..## $ $ #"
  "#.*##$$ $ #"
  "#. # $  $ #"
  "#.*# $ ####"
  "#...$@#####"
  "#  .#######"
  "###########")

(add-solution
  "LuuRUUdRUdldlddLdlluuuuuruurrrrrrrddLLLrrruullDulllllddldddddrruruurur"
  "uRldldlddrUrUrUUlDlDurrdLdLdlldlluuurUUddlddRUUUluuurrrrrrDDldddldlLdl"
  "UUUddrrruruulDlDurrdLdLLdlluurDrrruulDrdLLrruruuuruulllllldddddddRUrrr"
  "uruuuruurrddddLLLulDrdLulDrdLLLdlUUUUUddddrrrrururrruuLDrdLLLulDrdLulD"
  "rdLLLdlUUUUdddrrrrururrruuuulllDDuulldRDDDrdLulDrdLLLdlUUUddrrrruuuuuu"
  "rrdLulDDDDrdLulDrdLLLdlUUdrrrrururrruuuLLLulDDDDrdLulDrdLLLdlUrrrrurur"
  "ruuLLulDDDrdLulDrdLLLrrruuuurrDLulDDrdLulDrdLLrruruurrdLLulDrdLulDrdL")

(defmaze
  "###########"
  "#  ....#  #"
  "##$*.*.*.##"
  "#   ##..  #"
  "#  ##   $ #"
  "#    ##$$##"
  "# $$   $@##"
  "#  # ###$ #"
  "# $#  #   #"
  "#  #  #   #"
  "###########")

(add-solution
  "LUUluuLulllDDrUluRddllddrUUddlddrUUrrdLulluurruuRRdrddrddLLLrrruuluuRD"
  "DurrdLullulullddllddrrrdrrrrUUruLdddllllullluurruurrdrddRDrUlluurDDrru"
  "LLrdddDrddlUUUUUluullullddllddrrrdrrRlllullluurDuruurrdLullDDrUluRRdrr"
  "rddlUrdrruLLrddDDrddlUUUUUluuLLLullddrUluRddllddRUdlddrURRRRRlllllluur"
  "DldRuuUrurrrrddrdDlllluLrdrrrruuluulllldldDuuruulDDrurrrrddlUruLLLrrrd"
  "drruLLdlUrrdddDrddlUUUUdllllullluurrUdlDldRlddrURRRRRUrUruLLuLLLrrrdrd"
  "dlUlUdrruLuLLrrddlUruLrdrddDDrddlUUUUUruLdlUdddlllllluurDldRRRRRRurDDr"
  "ddlUUUUUruLLdlUrdddllluLdlluurDldRRRRRRurDDrddlUUUUUruLLdddllllldlddrU"
  "UluRRRRRRurDDrddlUUUUUruL")

(defmaze
  "###########"
  "#  #  # . #"
  "#  $  **. #"
  "## #$*... #"
  "#@$  ##.. #"
  "###   $ $ #"
  "#   $$ # ##"
  "#   #  #  #"
  "###########")

(add-solution
  "RuuRRDRRRdrUlllluurDldRRRdrruLLrruullDurrddlUddldLLLUUruRRDulldRRllldd"
  "rrrUrruuLDLddlddlULLdlluRRRRluuuruulDDDDuuullddRluurrrrRdddlLdLLURRlUU"
  "ruulDDDuullddRddrruLUUrurrdddLLdlUUlluurRRlllddrrUruLdddrrruuuLrdddlll"
  "uurRlluurDrrDrrddLLLLddrUllUUlluuRRRRdrrrddllLdlluuUdlluurRRldRRRllldd"
  "dlUruuullddRdrUUdlluurRurDldRRlldddrruLdlUUUdlluurRurDldRldddldlluRdrU"
  "UruuullddRdrUUdlluurRurD")

(defmaze
  "###########"
  "##  #####@#"
  "## ...  $ #"
  "#.*.#.$$###"
  "##* ##    #"
  "##..# $   #"
  "# $$   # ##"
  "#  ##$ $ ##"
  "#  ##   ###"
  "###########")

(add-solution
  "dLLDuLLdRdRluurDldDrrddLdllURUUlDruUUluLLdddlDRRRuruuruLLrdddrrddLdllU"
  "RUUlDllllddrUluRRuuuLDDDruuuuulDDDrddRRurUUluLrdrddrrddLdllURUUlDlllld"
  "drUluRRuuluuurDrrdrruLLLLDDlDDrRRurUdrrddLdllURUUlDllllddrUluRuuruurrr"
  "rdLulllulDDDDDRRRurUUruLLLrrdddrUdrddLdllURUUlDllluuuuurDrrrdDrdrddLdl"
  "lUdrruruululuullllddddrrRddrruruuruLLrdddldlluullluuuurrrrrDlulllldddd"
  "lddrUUrrruRUUruLLLLulDDurrrrdddldddrruruuuLrdddldlluuurUUruLLLLulDrrrr"
  "dddlddRUUUUruLLLrrrddrdddLdlUUUUUruLLrrddrdLrddlluUUUruL")

(defmaze
  "###########"
  "######@ # #"
  "##.. $$$$ #"
  "# .*.#... #"
  "# .$.**.$ #"
  "# $  .  $ #"
  "##### ##$ #"
  "# $ $ #   #"
  "#     #  ##"
  "###########")

(add-solution
  "rDDrrddLLLLrrrruulllDuruulDLLLrddRdRRUUddlluRllLUrurrDrrrddLLULuRluurD"
  "lddrdrruuuLLDldLdRuuruulDlldlddRUlullddRUrrrrUrurrdLrddlULLdRluLLdRurr"
  "rruullulDllDurrrrrddlllldlLURRluurrrrrddlUdldLLuluurrrDrrddLrddlUUruul"
  "lulllddrdrRUUrrdLDlllulldRRllulldRRurrdRRULLLuurrDullddrdrruruLddllulu"
  "urrurDDrrddLrddlUUruullulllddrdrRUUrrdLDlUdllulLdRRuRRllluurrDRdLuuurD"
  "lddrdrruuuLLDldLdlUrddddlluRdrUUURuLddddlllluRRRdrUUUruuruulDrrrdLrddl"
  "ULrrdddlUUruLdLL")

(defmaze
  "###########"
  "#@$  .....#"
  "####$##**##"
  "# $  # *..#"
  "#  $ # $.##"
  "#  $ # $# #"
  "# $       #"
  "#  ## #   #"
  "###########")

(add-solution
  "RRRRRRlllDDDLDlluuRDrruLdlldRlddrURRUdllluurrDullddrUrrdLrRRRdrruLLLuu"
  "uRRdLDldRdrruLdlUlLLuluuruurrrDDlDDuuruulllddlddlluRdlddrUUrrdLrrRuuur"
  "uulllddDDuuuurrrddldddlLullluuRRllddrrrdrruuurrdLuuuRDlullldDlllddrrrd"
  "rrRUUUUruLLrddddddrruLdlUUUUUruLdddddlllulllddrURRRRRdrruLdlUUUUUddddl"
  "llllluuurrrDDldRRRRdrruLdlUUUUlDDrdLLLuuulllddRdRRRRRdrruLdlUUUrUUddll"
  "uRdrUdlddllluuullDDldRRRRRRdrruLdlUUUluRdddllllluuRDldRRRRRdrruLdlUUUd"
  "dlllluuurDDldRRRRdrruLdlUUluR")

(defmaze
  "###########"
  "###   #####"
  "### $ #####"
  "####  #   #"
  "##  $$    #"
  "# .. #$  ##"
  "# **.. ##@#"
  "## *$$### #"
  "#.*..   $ #"
  "# .$ ######"
  "###########")

(add-solution
  "ddLLLLLULUUrURRRurrdLulDllldlddrddLULrruUlDuulldRDrUrdddlUUluuuRRDLulD"
  "rruRRRurrdLLLLLdDrrUdllDlddrUrUdlluRuulDDruuurRddLLddrUluuurrddLruuRur"
  "rdLdLruulDllldlluRRdDrddlUllUdrUdrruLuuRuuulldRDDDDLdRddlUUruuurRRurrd"
  "LLLLulDDDrrUdlluurRlluuurDDldddrruuLrddlluUUrdrRurrdLLLLrddlluUrrrrdLu"
  "lllddrrUdllDlddrUrULuuuruuulldRDDDDDrdLuuuurRdrruLLLuluurDDlDDDDuuuRRd"
  "rruLLLulDD")

(defmaze
  "###########"
  "###   #####"
  "## $ $@####"
  "#  $ $#.###"
  "#  #    ###"
  "# $  #*.###"
  "# $ ##.$*.#"
  "# # ##. # #"
  "# #$##$$$$#"
  "#  .......#"
  "###########")

(add-solution
  "LDlLrdRuuullDlDRllddRUluRururrdddldLruruuulldlddldddddrrUUUdddlluuuRUl"
  "uurururrdddldLDlluuururRlldldddrruruUddldlldddrrrrrUUrUUULLrrdddlddlll"
  "lluuurrDDuuurUrrrdddlddllLuuulluuurururrDDuulldldldddrruruRldldlluRluu"
  "rDDuuururrdLrddlULrddLddddrrruuruuuLLrrdddlddrUdrrULLdlllluuuurUruuull"
  "DlDlddddddRRRRRRRllllllluuuRluuurururrdddldlDDDuulldddrRRRRRlllllluuuu"
  "RluurDDldRuuuururrdLrddldLDDDuulldddrRRRRUrUUddrrUdllldllllluuuuurDldR"
  "urDDDuulldddrRRRlluuuuruULrruullDlDDDldRurDDDuulldddrRRluuuuruuLuurDll"
  "DDDldRurDDDuulldddrRuuuururuLLulDDDldRurDDD")

(defmaze
  "###########"
  "##@##   ###"
  "##$##   ###"
  "## ###  ###"
  "#   ### *.#"
  "#  $$ #$*.#"
  "# $   #  .#"
  "# $$    #.#"
  "#  ## $***#"
  "##### ....#"
  "###########")

(add-solution
  "DDDlddRRllddrURRRddrrrrUUUUddddllURdllluRRdrrUUUllDDldRuuurrddLLdlluuR"
  "lddrruLrrrdLruuullDldRRdrUUdlluuUUUluurDDDDDDlllllluuurrDDuulldddrRRRR"
  "dRRdrUllUUUUUluurDDDDDDlluuLDldRRRddlUrdRRuLUUUUUluurDDDDDDDulLulldRRR"
  "lddrRuUUUUUluurDDDDDDlllllluRRlluurDDldRRRRRDRRlUUUUUluurDDDDDDDlluuul"
  "DldRRRdrUUUUUluurDDDDDDllllluuurDDldRRRRdRUUUUUluurDDDDDD")

(defmaze
  "##########"
  "#  #######"
  "#  #######"
  "#.*...   #"
  "#.*.*#$# #"
  "# #. # $ #"
  "#  *$# $ #"
  "#  $@#   #"
  "####$# $ #"
  "####     #"
  "##########")

(add-solution
  "LUUUddlluuRURRRRDDDDrruuuuLLLLLDllddrrUrULdddrUUluuRRRddddrrddllllUUlu"
  "lluuRluRluurDDldddrrdrddrrrruulluuuullLDDrDDlUlluuRUrrDLulldddrrrUdlll"
  "uuurrdDrdLuuulldRurrrrddddrrddllllUUluurUdlddrUUddddrrrruulluuuulLLDll"
  "dddRRlluuurrurrrddddrrddlllluUllluuuuRluurDDlddddrrrddrrrruulluuuullLd"
  "DrDDllluuuRUrrDLullddddrrruUluurrrddddrrddllllUUllluuuurrddrUlulldRluu"
  "urDDlddddrrUdlluRdrrUUddddrrrruulluuuulLLLulDrrrrrddddrruuuuLLLLLLrrrr"
  "rrdddLdlUUUddrruuulLLLLrrrrrddLdlUUdrruulLLLrrrrddddddllUUUUUdrruulLLr"
  "rrdddddLdlUUUUUdrruulL")

(defmaze
  "##########"
  "##      ##"
  "## $$$ $ #"
  "##    # .#"
  "###$$$##.#"
  "#@$ # ...#"
  "# $   # .#"
  "# $#$#...#"
  "# $  ... #"
  "#######  #"
  "##########")

(add-solution
  "dddRRRRRurrddlUUdlllllluuRDldRRRRRurruullldlLrrurrruuuLulllllddrDDuulu"
  "urrrrrdrdddddddlUlllllluuuRDlddrrrrrrruuullldlDuLUlldddrrRRRllllluuRDl"
  "dRRRuuLrrurrrddLUruLrddddlUUUruuuluulllllddrDDlDlddrrrRRllllluururuulu"
  "urrrrrdrdddddddlUlllllluurDldRRRRRRdrUUUUUddddlllluuLuuurrDLulDDlDDldR"
  "RRRRRdrUUUUdddlllluuLuuuluurDDDDlDDldRRRRRRdrUUUddlllluuLuuuuurDurDurr"
  "dLLLulDDDDlDDldRRRRRRdrUUdlllluuLuurrULulDDDlDDldRRRRRRdrUlllluuLuuruu"
  "urrdLLulDDDDlDDldRRRRRlluuLrrUUUluurrdLLulDDDDlDDldRRRRluuLulDDldRRR")

(defmaze
  "###########"
  "#..... ####"
  "#..$#*$####"
  "#..  $ ####"
  "## # #  ###"
  "## # $  ###"
  "## # $$  ##"
  "#  ### $$ #"
  "# $ @#    #"
  "##   ###  #"
  "###########")

(add-solution
  "dllUUUUUUluurrrrrDDLLLrrruulllDLDDDDDldRdrruLdlUUUUUUlURurrrrddllLrrru"
  "ulllldDrrddRRUUddDrrDrddlUllUUrDluUlluurrDrDLddrUrDlluuuulldddRRuLdlUr"
  "rddrrrddlUruLuLLddRluuuuulllluurrrrDDlllUluRRlddlDDDDldRdrruLdlUUUUUUl"
  "uRdrrrrDDllUdrruulLLUlDDDDDldRdrruLdlUUUUUUluRdrrrrddDrrdrddlUllUUrrDL"
  "ulUlldRluuurrDrDLLdlUUdrrrDLrrdrddlUllUUruLuulLLLuuRlddrrdddRluuurrddL"
  "dlUUdrruulLLUlDDDDDldRdrruLdlUUUUUUrrrrdrddrDrddlUlluUruLLdlUUdrruulLL"
  "uLDDDDDldRdrruLdlUUUUUUUrdrrrdddddrUdrruLLdlUUdrruLuLLdlUUdrruulLLulDD"
  "DDDldRdrruLdlUUUUUUrrrrdrdddlUruLLdlUUdrruulLL")

(defmaze
  "##########"
  "####  #  #"
  "### $  $ #"
  "# $. $ $ #"
  "# .. #   #"
  "##** #$$ #"
  "# .*#@$  #"
  "#..*.*$###"
  "# . .  ###"
  "##########")

(add-solution
  "RUUULLdllDDlddrrrrrUUddlluRdlllluurDRdrruLLLuuurrdLDldRuuulDrruuruulDl"
  "DurrdLrrdrruLruulDLrrddlUdlUrrdddLUruullllddlluRddddrruuRUrruullllulDr"
  "rrrrddlldlddlluuululuRRRllddrrULuurDldddddrruururruullLLulDDDuulldRurr"
  "rrrrddlldlddllulUUrrUdllddrdrruururruulllLuurDldddllURdrUURuulDlDDrUlu"
  "RddlldddrdrruururruulLLLrrrrddllUddlddrUUUdddllluRlluuurrdLDldRuuulDrr"
  "uuulDrddlluluRRlddrruLuurDldddddrrruuurruulLLLulDDDuulldRurrrrrrddlldd"
  "dlllulUUrrUdllddrdrrruuurruulllLuurDldddllURdrUURuulDlDDrUdlldddrdrrru"
  "uurruulDuLLLdlldddrdrrruuuUddddllluluuurrurRdrrddLruullulldlldddrdrrru"
  "UlDrdLruurruullulldlldddrRlluuurrdLDldRuuruuruulDlDDDDuuuurrdLulDDDulD"
  "DuuluRdrruulDrdLrurrdrruLLLLulDlDDrruLrurrrruulDLrrdLLLLulDDrdLruuuurD"
  "rdLLulDDrurrrrdddllUUruLLLulDrrrdddrUUruLLLL")

(defmaze
  "##########"
  "####  ####"
  "##.*    ##"
  "#  .* #$ #"
  "#$.**$#  #"
  "#.*. $@$ #"
  "#..*##$$ #"
  "#.$ ##   #"
  "### ###$ #"
  "###.###  #"
  "##########")

(add-solution
  "LUULuLrdrddLULUllDDuRDRDDDuLUUluurrDDurrdLLuulldRRdrruuLuurDrrDrddLLLu"
  "UddrrruululLdlLdlluRRddRRuuurrdrddlLLuLLuRdrdrrruulullulDDurrrdrddlllu"
  "UddrrruululLdllddRRlluullddRRlddrUUlluurDDuurrrddLLUdrruulLdlldRRddlUU"
  "ruurrddLruulldRlDlddrUUlluuRDDrruuLrddlluRdddlUUruurrurrdrddlLLLrrrruu"
  "lulldlldDuurruulDrrrdrddlllUUddlluuRuRDllddlddrUUlluuRDDrrrrrruulDrdLL"
  "LLULrUruulDDLdDlluuRlddrrrrULrdrruuuLLulDDurrrdrddddlUruLLLLLLrrrrrddl"
  "UruLLLLrrrrrddddlUUUruLLLLruUddrruuulLulD")

(defmaze
  "###############"
  "############  #"
  "##########  $ #"
  "#  .... . $ $ #"
  "#   #$***.$$  #"
  "# $##  .*# $$ #"
  "#  ##  ..  #@##"
  "###############")

(add-solution
  "ULUULLdRluLLDRluLLDRddrrrrUrrruuLLLLLdDLdllURuulDrddrRuuuLrdddlluuRDuu"
  "rDurrrrruulDrdLLLuRdrrddlllULLulDrrrdrrruullulDrrrddllUUlLdlluRRRRddrr"
  "uuLruulDrdddlluLLruuRDlddrrruuLLulDrrrddllUUlllldlluRRRRdLruRRddrruuLr"
  "uulDrdddlluLuuRDlddrrruuLLulDrrrddllUUllllllLLdlluRRRRRdddlUUruRRdLulD"
  "rruRRRddrruuLLrruulDrdddlluuLuRdddldLruULrruulDrddrruuLrddlluUlLdLLuRR"
  "RRdLLrrdrruLLrruLruulDDrddlluLuuRDlllllLLLdlluRRRRRRRRRddrrruulLulDrrr"
  "ddllUUlllllldRlddrUluulldllddrUUluRRRRRRRRRRddrruuLLLLLLLLLrrrrrrrrruu"
  "lDrdLLLLLLLLrrrrruRdrruulDrdLLLLLLLrrrrrddrUruLLLLLLrrrrddlUruLL")

(defmaze
  "#########"
  "# *.....#"
  "#  ##*. #"
  "# # .*$ #"
  "#   **  #"
  "##$ $   #"
  "#@$  $$$#"
  "##      #"
  "#########")

(add-solution
  "RdrrrrrUUluULDrdrddlUdlluRdllluRuuRRurrDDuulldDLullluuuRRRRDDrrddLULLd"
  "RuuuullldlddrDRddrrUUUdddlluuRlddrruLuurDurrdLulldlluRdrddrrUULrddlluR"
  "dlllURuulluururrrdDDDLruuuullldlddrrRdrrruuuuLLddDldRlullDRddrrULrdrrU"
  "LrUULLuuurrdDLdlLDlluluururRRdDrdDLLuRdrrddlUdllURuuurrDDuulluullldldd"
  "rRRDrUllldRurrdddllUUddrruLruulDlluRlluururrrdDrrddLLddlllUUruuRRuulll"
  "dlddRdddrrruurruuuuLLLLrrrrddddllUUUdddddrrUUUUULuLLrrdrdddddllllluurR"
  "UruUruLrdrdddddlluUUUUdddllluRRdrUUddddlUUlldRRdrUUUdrruLUUrdddddlUUUU")

(defmaze
  "###############"
  "#    #  #    ##"
  "#  #$$    $.*.#"
  "#  $  $$$ #.*.#"
  "# $ $ $    #..#"
  "##@###### ....#"
  "###############")

(add-solution
  "UUluurrrDRRlddRRRUruurrddRddlllUUdllllLLrruurDRdrruLLLullDRdRRuLdlUrrr"
  "rdLLuluurDldRdrrrdrruuluurDDurDlluullDRRllLLddrrUdllulldRRuuulDrrRurrd"
  "LLrrdrruLLulldddrdrrUUluulldLLddRluurrurrddrddlllUlluurrRurrDDuulldRll"
  "llddrrUdllllulLdRRluuullldddRRurrdRRuLruulDLrrrRurrdLLrrdrruLLulldddrd"
  "rrUUluulldLLddRluurrurrddrddlllUlluurrRurrDDuulldRllllddrrUdllllUlldRR"
  "RRuLLrruulDrrRurrdLLrrdrruLLulldddrdrrrUdlUUluulldLLddRluurrurrddrddll"
  "lUlluurrRurrDDDrdLLruuuulldRurDDDrdLuulullllddrUluRRRRurDrDDuulllllddl"
  "lluRdrrrUluRRRdRluurDrDulldRlullllulDrrddllllluRdllluuurrrDRRRRRRurDll"
  "ddllllUdlluRuRRRRRRRurDDDuulllddllllUluRRRRRRRurDDulllddlllUluRRRRRRur"
  "DlllddllUdlllllluRRRRRRluRRRRRRlllddlUluRRRRlllllllullDldRRRdrUluRRRRR"
  "R")

(defmaze
  "##############"
  "###.$  #     #"
  "## .    #$ $ #"
  "#..* # $     #"
  "#..* # #######"
  "#.#### $   ###"
  "#.#@ $$$ $ ###"
  "#.##   #   ###"
  "##############")

(add-solution
  "rdrrUUUURuLLLddLLulDDDuuruuRDrdLLruuuRDrrdddddlluRdrUUUUruLLLulDldlDDu"
  "rrrULulDrdLruuurDrrdddRRdLLdlluRdrUUUUruLLLulDrrruLdllldlDrrrULLruRRur"
  "DllluRdrrDDDrdLdlluRdrUUUUrRRRuurrddLLLLLuLLLulDDrdLLrruurrDDDrdLdlluR"
  "drUUUUrrrruurDrdLLLLLuLLLulDDrdLruurrDDDrrddrruuLLLdLUUUruLLLulDrdLrur"
  "rdddddlluRdrUUUUrrrruulDrdLLLuLLLulDDrurrDDDrrddrUruLLLdLUUUruLLLulDrr"
  "rdddddlluRdrUUUUruLLLrruLL")

(defmaze
  "#########"
  "#.*....##"
  "#  #**$.#"
  "# #@*. ##"
  "#  $#$$ #"
  "##  #   #"
  "##$$  $ #"
  "##      #"
  "#########")

(add-solution
  "RRDDrrddlUdlluRdlllURdrruuuullDlDururrddddlluUluluuuRRRRDDlUruLLLdlddr"
  "drddrruuUUddddrruuuLrdddlllluululuururrddrDrrdddlllllUUrddrruLLrrdrruL"
  "LLrrruulDululuurDRDDlUruLrddrddLUUdddlluRdlllURRdrrUruullullDDuurruull"
  "ldlddRluururrrdddrrddlUddlluRUddllluRRdrrruullUUllDDuurrddrrddllllluRu"
  "ulluururrrDDlldddlUruurruullldlddRluururrrddllDlddrdrrrruullUdrruLrddd"
  "llllluuuluururrrdrDLuullldlddrddrdrrrruullUUrDrdddllllluuuruRUdldlddrd"
  "rrrruuuluuLuLLrrdrddlUlUdrruLuLrdrddrdddllUUUUUdlUdldlddrRdrUUUUruLrdd"
  "rddLdlUUUUlldDldRRdrUUUruLrdrdLdlUU")

(defmaze
  "#############"
  "#   ....... #"
  "# $.######. #"
  "#.*.*      ##"
  "#########  ##"
  "#       # ###"
  "# $$$$$$  $@#"
  "# $ $   $ ###"
  "#       #####"
  "#############")

(add-solution
  "LLuuruuulllllllddRRRRRRdrUUruLLLLLLrrrrrddllllllluullddRRRRRRRRddddLLd"
  "lllllluuurrrrrrDRdrUUUrUUruLLLLLrrrrddLLLLLLLuullddRRRRRRRRdddllullDRR"
  "RdrUUUrUUruLLLLrrrddLLLLLLLuullddRRRRRRRRdddllllullDRRRRRdrUUUrUUruLLL"
  "rrddLLLLLLLuullddRRRRRRRRdrUUruLLrdddlddllllllulldRRRRRRRdrUUUruLLLLLL"
  "LuullddRRRRRRRRdddllllllllddrUluRRRRRRRdrUUUrUUruLddLLLLLLLuullddRRRRR"
  "RRRdddlllllllddrUluRRRRRRdrUUUrUUdLLLLLLLuulldRldRRRRRRRRdddllllllddrU"
  "luRRRRRdrUUUrULLLLLLLLrrrrrrrdddlllllddrUluRRRRdrUUUruLLLLLLLrrrrrrddd"
  "llllddrUluRRRdrUUUruLLLLLLrrrrrdddlllddrUluRRdrUUUruLLLLL")

(defmaze
  "#############"
  "#      ###.##"
  "# $$$ $###.##"
  "#     $ ##.##"
  "# $# $ * #.##"
  "#   $# ..**##"
  "#    ##. .*@#"
  "#############")

(add-solution
  "LUUUUdddLLuLuLLuullllddrDldRluuuurDDDuuurrrddLLrruulDrdLuLulldddddrrrU"
  "UUrrrdrddlUlULulDrrdrrrrdLuluLuLLdRurDldRRRdrUUUddlllluluuurDDldRurDld"
  "RRRdrUUdlllluluuulldRRurDDldRurDldRRRdrUllldRRllulullUruulldRurDDldRRu"
  "rDldRurDldRlululluululldRRRurDDlulllddrURRdRRurDDluuulDldRRurDldRRRlll"
  "ullullluRRRurDDllldlddrUUluRRRlldddrrUURRurDldRRlluuulDulldRllddRdrUUR"
  "RurDldRluuulDulldRdRRurDluulDldRR")

(defmaze
  "#########"
  "#   #####"
  "# $ $ . #"
  "# #  ** #"
  "#  $# ..#"
  "#    .* #"
  "# $$##**#"
  "###  @. #"
  "#########")

(add-solution
  "rrUULuULrddrddlUULuurrDDllLLUURRddrddlllUdrrruuluulldllddRURuurrddrruL"
  "dlLLuurrDrdddlllUUrRurDllldlluRRRRuullDurrddlLrrurrdLulullddddrrrUUruu"
  "LLrrddlddllluurRuurrdLDlllddrrrrUUlllldlluuuuurrDDDDrrruuluLrdrruLLrdd"
  "dllluuRRurDlllddrrUdlluurRddrrddllllUlluuRDRRRuullDurrurrdLLddlLddrrrr"
  "uuLUdrddlUUddllluurRuuurrdDLDuruulDrdDllllddrrrrUUllllulluuurrDDRRuLdl"
  "DlluuurrDrrrrdLLullulldddrrDrrrUdrUdllllulldRRRRUrrdddllllUdrrrruuulld"
  "lLullddRUrrrurrdddllllUdrrrruuulldRurDlDullLdlluRRRRurDllllluuuurrdrrr"
  "DurDlulDullDDuRRlluulldddRdRRRUdlllluuuRRRRldlDldRR")

(defmaze
  "##########"
  "#  #     #"
  "#  $   $ #"
  "# $#$##$##"
  "##@$     #"
  "###$#*#  #"
  "#...*.*. #"
  "## #$.$*##"
  "##. ..  ##"
  "##########")

(add-solution
  "RRRDDLLLddrrrrrUULDrdLLLLrrrruuruulDDlLLDurDuuulllUluurDRRurrdLrurrdLD"
  "DDuuuulldRurDDDrddLUUUdddLdlluRuuRllllUluurDrRurrdLrurrdLdDrddlLrruulD"
  "uuuulldRurDDDlllllUluurDrRRurrdddlllUdrrruuulldRurDllLddllUluurDRRRRll"
  "ddllUluRRRddrdDlLUdrruurrrddLLLuuluullddRRuuurrdLLLrrrurrdLLLLddRddrrr"
  "uulUdrddllddrUluluullddRdRUlluuruurrurrdLddrddLLLLrrrruulDrdLLLrruuuuu"
  "lldRurDDDDrdLLrdLruuuuulllddrRlluurrrdDDrdLuullllluluurDRRRRurDDDDDuul"
  "llllUluRRRRRurDDDD")

(defmaze
  "###########"
  "#  ####   #"
  "#.. $ # $ #"
  "#.$   $   #"
  "#.*# $$ $ #"
  "#...# $$$ #"
  "# ... #@###"
  "###########")

(add-solution
  "UUULLuLLLulDDRDDlUUrurrrdDlULulDrrruLdlllddrrdrrUUlulluulDDrrruLdlluur"
  "DrrddrddllullUURurrrDrrrruullDurrddlLLLullldlddrrdrruUlURdddllulluurRl"
  "lddrUluRuulDDrddrdrruuuuLLrrddddlluluuluurDDurrrddddllullUUrurrrdRRddL"
  "ruuuurrddLrddLLuUllullldlddrrdrrUUlURRRlldddllulluurRllddrUluRuulDDrdd"
  "rdrruuuuLLrrddddlluluuluurDDurrrddddllullUUrurrrdrrddLruullullldlddrrd"
  "rrUdllulluurDldRldRRluuuurrrdrrddllUlULulDDDldRuuurrdrdrruulLLLulDDDuu"
  "rrrrrddrruuLLLLLLulDDuuulDDDrurrrrrddrUruLLLLLLLulDrrrrrrrruulDrdLLLLL"
  "LLrrrrrruulDrdLLLLLrruLLLDrrrddrUruLLdlUruLL")

(defmaze
  "#########"
  "# ... ###"
  "# . # $ #"
  "##.# $  #"
  "# .$$@# #"
  "#$. $$$ #"
  "#..*  $ #"
  "#  .$ ###"
  "#########")

(add-solution
  "UrrdddLLULLrrdrruuulllDurrrdddlluUluRddddLLUlldRuUUlDrdrRUruulDLDurrdL"
  "dLruulDlluRRdrruulDrdLddrUUlulldRurDrdrruLLLuurDldRddlUULulUUrurrDDDLd"
  "drUluuruulldlddldRRluuururrddldDrdrruLLLLruurDuuulldlddRdrRdrruuuuLrdd"
  "ddllulluluururrDrrddddlldlUrrruuuullDDDuuuulldlddrdRluluururrdddLddRUl"
  "uuruulldlddldRRluuururrddldDlluuuluRRdldddrruLdlUUUluRdddrrururrdddLLL"
  "rddlUUrrrdLLulLulldRRRRULLdlUUUddrrrdddlUUruLLdlUUdrrrddrruLLdlUrrruuu"
  "llDDLLdlUrrddrruLdlUrrruuLulDDLLrddrruLLL")

(defmaze
  "##########"
  "#  #     #"
  "#  #$$@$ #"
  "#  # .#$ #"
  "## # ..$ #"
  "## # * $ #"
  "#  # *#*##"
  "# $$ .*.##"
  "## ...  ##"
  "##########")

(add-solution
  "urrdddLLdLUrrrdLulldDlddrrruUULullDDrUrrdddlllUdrrruuulldDRdrUllldllUR"
  "RRRdLulllUUUUluurDDDDDldRRRRuulDrdLLdlUrrrdLruuurrruuuullllDDDDDruruLr"
  "drruuuulldLrurrdLrddLUddlluRdrruuuullldDlDRddlUrdddlUllUUUUluurDDDDDld"
  "RRRdrUllldRurrUUlDrurrruuuulllddDulDDRddLLdlUrrrdLullUUUUluurDDDDDldRR"
  "RdrUllldRurrUUlDrdLLdlUrrrdLuuuuurDDDDuuuuuurrrddddlLrruuuullllDDDDDDl"
  "lUUUUluurDDDDDldRRRuuurDrrruulDLrruuulDLulDrrDrdLrdLL")

(defmaze
  "########"
  "#  #  ##"
  "#..** ##"
  "# .#. ##"
  "# #.* ##"
  "# $   ##"
  "# $ $  #"
  "#$$##  #"
  "#  @#  #"
  "########")

(add-solution
  "llUUURRuRuruulDrdDlldllUUruRRDrdLrDDrddlUUULuuulldlddrRdRUrDrddlUruLul"
  "llluururrdrDDuuluurDDlulldlddrrrUUruLdddrUUddDrddlUUUlldRlulldRlddrUUl"
  "uuuruulDDDDDrRuruuuLrdddldlluuuuurDrrdddrDLLruuuullldddRRdrUrdrddlUruL"
  "ulllluuurrrdrDDuuluurDDlullldddrrrUUruLLrdddrUUddDrddlUUUlldRlulldRddl"
  "UUUUUddrrrrDlLullddrUluRRdrrrddlUruLLLulDrrruLuurDDDluuuuurDDDlddlluRd"
  "rUUUruLddddlllddrUUluRRdrUUUdddrrddlUruLLLullddrUluRRdrUUrDDrddlUruLLL"
  "ullddrUluRRdrU")

(defmaze
  "###########"
  "#  ########"
  "# $####  ##"
  "#      $ ##"
  "## .*$# $ #"
  "## $*.$ #@#"
  "###...$ ###"
  "###.***   #"
  "####..# $ #"
  "####  #   #"
  "###########")

(add-solution
  "uLuLLLrrDDDDrrddllUUULLDRlLddrUluRurrUUruulDllDDDDurruUruLdddlluRluLLD"
  "RRddlUrrrULrdDrrddllUULulldRRllddrUluurrdLrRddrruuLLLuurDldRlullURdlDR"
  "ddlUUrurrdLrddrruuLLuuUruulDDDDlluuuRlddlluuRRdddDlUruuullddRluurrdLdd"
  "rUdrrDLruuuruulDDDLdlluuurRlldddrruLdlUrrruruulDllDDrrUruLddlddrUUllld"
  "RldRlddrUUluurrrddLLrrrrddllUdrruulLuuUruulDDDlluuRllllluurDldRRRRddld"
  "DrddlUUrUUrrDDLrrrddllUdrruulLuuuruulDDDlllllUUrrrRllllddrrrdDlUruLddd"
  "drUUluurrrDDLLUlDrrrrrddllUdrruulLLruuuruulDDDDuuulllllluurDldRRRRRllD"
  "DRdDlUUrddrRddrruuLLLLrrddrUruLLLruLLrruuruulDllDDuurrDDDDrddlUruLuuLr"
  "uulllllddR")

(defmaze
  "##########"
  "#####    #"
  "#   $$   #"
  "# ..#    #"
  "##$.#$#$##"
  "# $.#  $ #"
  "#  .*.$$@#"
  "###....$ #"
  "##########")

(add-solution
  "uLUUruLdddrddLUUUUllDDRluurrruullDulDLLdlluRRddddlluRUruulDruRRurrrddl"
  "dDlluUddrruuLrruullDLrurrdLdddlluURlddrrrddlLLULLdRRuUrrrddlULLdRurruL"
  "uUllDDRluurrruullDulDLLdlluRRRRurrrddldDlddllluUUUddddrrruuruuruulDDuu"
  "lldRllLdddddrrruuluUddrddlUUrrrdLullddlluuuuurRurrrddlDDuuruulDDuulldR"
  "llldddddrruuUUddddlluuuuurRddddLrurrrddLUruLdLLuuuuurrrddlDDrdLuuuruul"
  "llddddRluuuurrrddlLrruullDllldlluRRRRurrrddldDllddllUUUdddrruLdlUUlUdl"
  "dRRdrruLdlUrrurrrddlLLLruurrDrdLLLrruuuuruulDDDDuuuulldDDDDurruuuLulDD"
  "DDrDLLrrurrdLLLrruuuLulDDD")

(defmaze
  "###############"
  "#   ###  ..   #"
  "#  $# $ ##.*. #"
  "#     $ $ $**.#"
  "#  $ ##$$ ...*#"
  "###  #@ ###   #"
  "###############")

(add-solution
  "rURRRRdrrUdllullllULLLLrddlUlUdrruLdlluuurrDDuullddRdRUrrrrdrruLLrrdrr"
  "drruLLrUULDLLdlluRRRRuulDrdLrurruLrdDLddrUUlulldRlLdlluRRRurrDLrrddllU"
  "LLrrdrruulullDurrdrddlluUllllLLLLdlluRRRRRRRRRuulllDurrrddllLLLLLrruRd"
  "rrrruulllDurrrddRddrruuLuuLLdRDrrddllUUluurrrDDlllllLulldRRRRRRRllllll"
  "lllddrUluRRRRRRRRllllllldllluRRRRRRRRdRUUdllllllllluurDldRRRRRRRRdrUll"
  "lllllluurDldRRRRRRRRllldRR")

(defmaze
  "###############"
  "#    # @# ....#"
  "#  $ $$$  ##..#"
  "# $ $   $$$ **#"
  "##   ##    ...#"
  "###############")

(add-solution
  "lDLdLdllUluurrrDrrRddrrULdlULullulllddrdrrUdlluRlluurrDRlullddrdrUdrUR"
  "RRdrruLLLuurDldLLdlluluurrrDRlulllddrdrrUdlluRlluurrDRlullddrdrUdrUrrR"
  "drruLLrrdrruLLrrdrrULuulllDurrrdddlluRdlllluuRDRdrruLLrrdrruLuulllDlld"
  "dRRRllluurrDrdLuuurrrrDlddLruuulllddrRdrUUruLLrddlllulldRRRRdrUUruLddl"
  "llllLLLdlluRRRRRRRRRdrUUddrUUdlllllllllllluurDldRRRRRRRRRRRdrUllllllll"
  "llluurDldRRRRRRRRRRllllllllluurDldRRRRRRRRlldRRRUllllluulDLdRRdRRRRllu"
  "ulDldRRRlllulllluurDldRRRRRRRR")

(defmaze
  "############"
  "#    #######"
  "#*$ $##  ###"
  "#.#..#@$ ###"
  "#.*.$##$ ###"
  "##..*.* $###"
  "## * ..$ ###"
  "## . ### $ #"
  "#### $ $ $ #"
  "####       #"
  "############")

(add-solution
  "RDrDDLLLuLLuLrdDlUrrrdrrruulDLdlLuRdldddrrrruUULLLulDulldRurrdrruLLLLd"
  "RurrrdrdddlllluUUluRRlddddrrrruuullLrrrdddlllluuUrrruurDDllluRdllluRRl"
  "luurDluuullDurrdLrurDDldddrruLdlUUUrDlddrruLdlUUrdddddrruLdlUUUruLdddr"
  "rdrruLLLrrrdrruLLLLdllUUUrrrrDuuuluurDDDDlUUddlluRdlllURRdldddrruLrrdr"
  "ruLLLdllUUUrrrrDulllldddrrrrruLuuuulDrddddlllluuuruRdllluRuuuulldRurDD"
  "DDRdLdddrruLdlUUUUUddddrrdrruLLLdlUUUruLdddrrrruuLLrruulDLrrdddrruLdlU"
  "ULuuruulDDrDLrDDrdLLLLdlUUU")

(defmaze
  "#########"
  "# $ *.  #"
  "#....$# #"
  "#..#    #"
  "#*. #$  #"
  "#. $    #"
  "##$$$$$ #"
  "##@#    #"
  "#########")

(add-solution
  "UUUUluRRuRRDulldRdRDrrdddlUdllURdrruuullulullddddRlulUrrDluuurrdrdDLLr"
  "rurrdLrddlUUruuuuLLDlllddrdrRlluluurrrurrddddddlllURuuUluluRdllluRdddr"
  "dRdrUllulldRdRRuLrddrrruuLLLdllUluUrurrdrDululldlddrUddrruLrddrrruulLu"
  "ulullulDDurrrdrddrrddlUdlluluulUluRddrddrRuuuluuRDDDDLddrrruuuuLrddddl"
  "llullUUrDldRluuUluurDDDrdrruUluuLrddrddlluluuRurrDDrrddLLLrrrddlllUllU"
  "UUluRluRRlddddlUUrdrDRddrrruulLLddrUluLulldRdRRuLdlUUUdrdrddrrruulLLLd"
  "lUUrdrrrruuuuLrddlDrdLLLLdlUruLdlUrrdrrruulDrdLLLdlUruLrdrddrrUruLLLLd"
  "lUrrddrUruLLLL")

(defmaze
  "###########"
  "##  ##   @#"
  "##  $ $ ###"
  "##  ##    #"
  "### ## #$.#"
  "### ## ...#"
  "### ###.$.#"
  "# $ ### ###"
  "#     $  ##"
  "###  #   ##"
  "###########")

(add-solution
  "lllDDrrrddlldddLLLdlUUUUUUluurDRRRdrrrddlldddlllluuuuUluRddddddlluRdrU"
  "UUUUluurDDDDDDldRRRRdrruLuuurruulllulLrrdrrrddlldddLLLLulldRRRRRdrUlll"
  "luuuuuluurDDDDDDldRRRRdrruLLLLLuuuuuurrrdrrrddlldDuurruullluurDllllddd"
  "dddrrrdrruLuuulUUddrdddLLLdlUUlldRRRRRdrUlllluUUUUluurDDDDDDldRRRRdrru"
  "LLLLLuuuuuurrRdddrdDuurrdLulluuulllddddddrrrdrruLuUddLLLdlUUUUUUluurDr"
  "rrdddRdrruuulLrrdddlluRdrUdllddllllulldRRRRRdrUlllluuuuUluurDDDDDDldRR"
  "RRdrruLLLLLuuuuuurrrDrrDDldDuuruulluurDllllddddddrrrdrruLuuulUUddrdddL"
  "LLdlUUlldRRRRRdrUUUddlllluUUUUluurDDDDDDuuuuurrRdddRRuuLrddldddlllLuuu"
  "uuurrrurDllllddddddrrrruuuluUddrruuLuLLLrrrdrddlluUddrdddllllulldRRRRR"
  "drUlllluuuuuluurDDDDDDldRRRRdrruLUURuuuluLLLrrdddRdddLLLLulldRRRRRdrUU"
  "UddlllluuuuuluurDDDDDDldRRRRdrUU")

(defmaze
  "###########"
  "#     #####"
  "# $  $ #  #"
  "####    $ #"
  "##### $$# #"
  "#####   $ #"
  "##  ### . #"
  "## $# .*..#"
  "##.*.$*$  #"
  "#@$ ....###"
  "###########")

(add-solution
  "RRRRRURUrruuuuLLLuLulldRdRRRRurDDDLLUdrddlUUdddldlluRuRlddrUrUrrdLLLdl"
  "luRuRRdLulDrrrruLLruurDuuulLLuLulldRdRddRRDrruuullLuLDrrrrdddlluUddrdd"
  "lUUddlluRdrruulDruruuulLLululldRdRRRdDllURuuLDrddrDrruLdddrUdlLUlldRRu"
  "ruuruuLLLddRDrdrUdlLdlluRRuuluurDDDrrULdlUrrdddLUdLUlldRRurruullDurrdd"
  "lLdllldlluRRRuRRdLdLLuRdrUlllldRRRllluuurDDldRRurrruuurrdddLLLLLdlluur"
  "DldRurrrruuuluurDDDDuurrdddlLLLLLrrrruuuuurrDDllluuululldRdRRlluurDldR"
  "urDRddrruuLLdddddrruUUllllURlUluurDrDrDDDDuurrdddlLLLLrrruuulluRuulldR"
  "urDrDDDDuurrdddlLLLrruuulluRlulululldRRRurDldRurDrDDDDuurrdddlLuuulluR"
  "urDDDuuurrDDD")

(defmaze
  "###########"
  "## .*.#####"
  "## .$.$ ###"
  "###.....$ #"
  "#  .#*### #"
  "# $   $@# #"
  "### ##$$$ #"
  "###  #    #"
  "###  # $$ #"
  "###  #    #"
  "###########")

(add-solution
  "LDDrrruuuuLLLLLruuLDlluRRdrdrruLLLulldRDRRlluurDrrrdrrddddllluuLUUddrd"
  "drrruuuullulLdddrrDldRluuluuuLulldRdDDDDrddlUUUUUlldRurDRRuuuLulldRDRl"
  "uurDrrrdrrdddLLuLLLLUUddrrrrdrruuullullDRlululldRRldddDDrddlUUUUUURRlu"
  "ulldRurDlddlldRurDDDrddlUUUUrrrrddlUruLLLLUURuulldRurDlddlldRurDDDrddl"
  "UUUUrrrrdrrdLrddlllUUUruLLLLUUUluRdddlldRurDDDrddlUUUUrrrddddrrULdlUUU"
  "ruLLLLUUUddlldRurDDDrddlUUUUrrrdddrrULdlUUruLLLLUUdlldRurDDDrddlUUUUrr"
  "rrddrruLLdlUruLLLLUlldRurDDDrddlUUUU")

(defmaze
  "################"
  "######  ###    #"
  "###.##    $$   #"
  "###.#  $$ #   ##"
  "###.#   # $$$###"
  "#..... #     $ #"
  "#   ####   ###@#"
  "################")

(add-solution
  "uLLLLdlUrrrUUlDrdLLdlluRURRllddrUruurruullDRlLLDDRdrrUUdLdldlluRuuurrD"
  "rddlLdlUrrruulDrdLLdlluRURRllddrUruuullDDuuLLulDDlddrUrURuLddlluRRuulD"
  "DrdLulDrdLLLdlluRdrUUUddrrruruurrrrurDurrdLLLLLLLulDDrdLulDrdLLLdlluRd"
  "rUUdrrruruurrrrdddldlluRuULLulDrdLulDrdLLLdlluRdrUrrrururuRRRdddLdlUUU"
  "ddrruuulLLLulDDrdLulDrdLLLLrrrrururrddrruLdlUUdrruulLLLulDDrdLulDrdLLL"
  "rrrururrurrurrddLruLLLLLLulDDrdLulDrdLLrrururrdrrUruLLLLLulDDrdLulDrdL"
  "rururrdrrdrUUruLLLLLLulDDrdLulD")

(defmaze
  "###########"
  "### .. ####"
  "###$*.*.  #"
  "#  $@#*.* #"
  "# $$# ..###"
  "# # $ $.  #"
  "# $   $*  #"
  "#       ###"
  "###########")

(add-solution
  "LDDRdrdrrULURuuurrdLulLulllDDrUluRdRRRDDDlddlluRlulUUrurrrdddlLddllUdl"
  "luRRdrruurruuullldldDRRddlULrrdrrULLdlluuuururrrddDuuullldlddddrruuuRl"
  "dddlluRluuuruulDDrurrrdrruLLLLLdlDDRddrruUrUdlddlluuRlddrruLuuRDllluur"
  "urrDurrdLullldlddrrrrUUddllddrrUUdrruLLddlluRRllulluururrRdDDLLddllURR"
  "dlllluRRdrruurruuullldldDRRddlULrrdrrULLrrUrrdLLdlluuRRddlllluuuuruulD"
  "DrurrrdrruLLLLLdlDDRRddrruuUdddlluuuRldlluururrDRlulldlddrrddrruuLUdrd"
  "rruLLddlluRRuUllDurrddlUllluururrrDDDlddllURdllUdlluRRdrruurruuullldld"
  "DRRddrrULdllULuuururrrddDuuullldldddrdrruLuuRUruLddDrUdllddlluRluuuruu"
  "lDDDDRRuruuLrddldlluurUrrddrdLuuurDurrdLLullldlddrddrrUUrUUddlddlluRlu"
  "luururrRdddlLddllUdlluRRdrruurruuullldldDRRddlULrrdrrULLrrUrrdLLdllllu"
  "uuururrrdDDuuullldlddddlluuuuRlddddrrrruuRRuuullllDrurrrdddlluRUruLLrd"
  "ddlddrUUUUruLrdrruLdllddrUdlddllluRRdrUUUURddlddrUUUddrruLdlUddllllllu"
  "RRlluuurrDDRRRddllUdlluRRRdrUUUddllllluuRurDDRRddllluRRRRU")

(defmaze
  "#########"
  "#       #"
  "# $ $ $ #"
  "##### ###"
  "#@### ###"
  "#$###   #"
  "# $   $ #"
  "#.**  $ #"
  "#..# $  #"
  "#.** #$ #"
  "#..... ##"
  "#########")

(add-solution
  "DDRRRddddrrUruuLLrruullDlDRlulllDDrURurrurrddddldllllUUluuRRllddrddrrr"
  "ruruuuullddLDDLdRuuuruurrddLrddlULrruuulldDlDRuuurrddLruulldDldDldllUR"
  "uURRddLruulldDldRuuulDrurrRuurrddLLLLuRdddlluluuRRlDrrrrrddlULrruLLLLL"
  "DDrrUruLLulDDlDRuuulDDrurrrrruulDrdLLLLLulDrrrrruulDrdLLLLrrruuuuuulld"
  "RurDDDDDrdLLLulDDurrruuuuullulldRRRurDDDDDrdLLLulDrrruuuuuurrdLulDDDDD"
  "rdLLLuRdrruulDrdLL")

(defmaze
  "################"
  "#         ######"
  "# $ $ $ #.*... #"
  "#   #### .#.*# #"
  "## $####$#..*..#"
  "## $    $ $ * ##"
  "##$###    $ ####"
  "##@###  ########"
  "################")

(add-solution
  "UURUUlluurrrrrrrrdRRddRRllddLLuRdlLLuRdrruuruulldlDDuuruullllllllddrrd"
  "dRRRRdRUllllluulluurrrrrrrrdrrddlddLruuruullullllllllddrrddrrrrrRdrrUU"
  "rrdLullDllllllluulluurrrrrrrrdrrRDulllullllllllddrrddrrrrrrrurRuulllul"
  "lllllDDlluRurrrrrrrdrrrddlldllllllllUUddrrrrrddlUrurrdLLrrururrdLdLLuR"
  "drUlllldRRurrruuullldlDDrruRUruLdddrUdlllluuruullllllllddrddrrrRRRRRur"
  "uuLrddldlllllllluuluurrrrrrrrDrrddldlllddlUrurrdLrururrdLdLLuRuruurDlu"
  "llullllllllddrddrrrRRRRdrrUUUruLLrddddlluRdrUUUruLdddrUUddlllllllllluu"
  "rDldRRRRRRRRdrUluRdrUdllllllddrUluRRRRdrUluRddllLullddrUluRRRRdrUUUddl"
  "lllllllluuluurrDDDldRRRRRRRdrUllllllluuuurrdLulDDDldRRRRRRRRRllllllllu"
  "uurrurrdLLLulDDDldRRRRRRRRdrUUdllllllllluuluRurDDDldRRRRRRRRdrU")

(defmaze
  "#############"
  "#   #### ...#"
  "#  ..*.*..$ #"
  "# $$.#.   ###"
  "#   ## $ ####"
  "#######$$$ ##"
  "#         $@#"
  "# $ #####$ ##"
  "#   #####  ##"
  "#############")

(add-solution
  "LLLUUUlluLLLulldRlddrURlluurDurDrrrdrrruulDrdLruurrdLullddrUluRRdllLLL"
  "LLulldRRRRRRRddddrruLdlUUUrULuRddlddrrdddlUUruLdlLLLLLddlluuRRRRRRRUrD"
  "luUUruulDLLLLLulldRRRRRRRDDurUlllddRDRdLuuluulllllddrUluRRRRRRRdDlluul"
  "lllddrUluRRRRddrruuLLLLrrrrdLdddrUdrruLLdlUUUruLLLrrdddrrrdddlUUruLLdl"
  "UUUruLLrddddLLLLddlluuRRRRRRRurrddlUruLLdlUUUruLddddllllddlUluRRRRRRRu"
  "rrddlUruLLUUdddlUUU")

(defmaze
  "#########"
  "##      #"
  "# $ $$$ #"
  "#. ##@$ #"
  "#. # $  #"
  "#.# $ $ #"
  "#.*.$  ##"
  "#.*... ##"
  "#########")

(add-solution
  "DrruuullDLullDldddddRRuRRUdlluRuRUdlddrUdlldlluRRurrddLLulluuuururrrrr"
  "ddddLruuLDruuullllldlddddrruruRDlldlluuuururrrrrddddLruLruuullllldlddd"
  "drrdrrrUUddllluurDllluuuururrrrrddddldLruruulDDruuuullllldlddddrruruRl"
  "dldlluuuururrrrrddddlLdLLrruulDrrruLdlddrUUllldRldRuurrruuuullllldlddd"
  "ddRluuuuururrrrrddddllldlLdlUrrrdLruurrdLrdLruuruulDDruuuulDDuulldRlul"
  "ldRRlllddrUluRurrrrrddddllddrUlulldllUUruuurrrrrddddLruulDruuulllllddd"
  "lddrruruRDllddRUllldRluuuruuurrrrrddddLdLLLLdlUUdrrrrruruulDlDurrdLdLL"
  "LLdlUrrrdLLrrurrdLLLrrruuulDrdLLLLrrrruruuuulDulldRDDDDrdLLrruuruuuLul"
  "DDDDDuuuululldRRurDDDDrdLLLrrruruuLulDDDrdLLruulDD")

(defmaze
  "############"
  "# $   .....#"
  "#@$$$ $* ###"
  "##$  ##.####"
  "#  $## .. ##"
  "#    # .#$ #"
  "# $$ ##.*  #"
  "#    ##    #"
  "############")

(add-solution
  "uRRRRRRRRllDulldRlullDRlulldRDDldRlddrUUUdddrruuLrdLdlluuurUUluRRRdLul"
  "DDurrdLruuRRdLruRRdLDDrrDrddllluUUUUruLLrddddddrrULdlUUUUULuLLLddrURRu"
  "RDDDrrDrdLdlluUluRuulllluRRdlllluRRddDDrddllluuuRUruulldRdrruruLrdrruL"
  "dlldlluRRRRurDDDDuuullldlldldddrrruulUUluRRRRurDDDuullldlddrddlUUUUluR"
  "RRRurDDullldldllddRdrUUUUluRRRRurDllluRRdlllluRRRdldldlldRdrUUUluRRRR")

(defmaze
  "################"
  "#######@########"
  "#  ####$#  #####"
  "# $#### $. #####"
  "#      .#* #####"
  "###### *...*   #"
  "####  *.** $$  #"
  "#### $ ##.$  $ #"
  "####   ##.  #  #"
  "################")

(add-solution
  "DDRRurDDDlDDRUlulLrrUruulDDDDrURRDrruLLLLdRlddrUUluUluurDDDDrddlUUrRur"
  "rdLrddlULrruulDulDlllUlldRRRlllLddlluuRRRurrdLruUruulDllDlDRRdrrUUddll"
  "ullDllddrrUURurRdLLrrrRRurrdLLLLLulluurrDrDLdRRRuLLdddrUUrurrdLrddlULr"
  "ruulDulDlluuluurDDlulldddLrRRUUruLdddRUUddRRurrdLrddlUULuLLdRlddrURllu"
  "urrDullddrUluuluurDDDDlUUrddrruLLdllllddlluuRRRuulDrdRRRuulDruuulDrddl"
  "LdlLuuruRRDrdLrrrdrruLLrrdddlUUluLLdRdRluurDrruLLdlluuluurDDlulldlddrr"
  "RUUruLdddRUUddRddlUUrRurrdLrddlULrruulDulDllluurDldRRRuLLdddrUUluuluur"
  "DDDDlUrdrruLLrdddlUUllllddlluuRRRRRRdrUrurrdLrddlULLrrruulDulDlllllluu"
  "ruRldlllllluurDldRRRRRddddlUluRRRRRRRRurrdLLLLLLLrrrrddlUruLLLrrrrrrdd"
  "lULrruLLLLLrddrUruLL")

(defmaze
  "################"
  "##@####  #######"
  "# $   #  #######"
  "#  $$$# . ##   #"
  "# $  .##*$#    #"
  "####......  #$ #"
  "########.#$    #"
  "########. $ $$ #"
  "########.*  #  #"
  "################")

(add-solution
  "DrrrDDldRRRRRRllddrRUrrrUrdddlUUruuulldlDLdRlddrURlluurDldRuuururrdddl"
  "LrruuulldldDldlluuRRllUUrDldRluuluurDDrddlUdddrrurrrruuulldlDLLrddlluU"
  "rrrururrdddllLulllddrrdrUllldRluuurrDurururrdddllLdLLdlUrrrdLLrruuulll"
  "llluuullllddRRRllluurrrrdDldRRRRRRDDrdLruurrruuulldlDururrdddllLulllll"
  "luuullDRurDDldRRRRRRDrdLLrrurrruuulldlDururrdddllLullllllluululldRRRur"
  "DDlulldRRdRRRRRRDrddlUUrrrruuulldlDLLLLLLrrrrrrururrdddddlULLdlUUdrruL"
  "uLLLLLrrrrrddlUruLLLLrrrrdrdrruLLLdlUruLLLrrrdrrruuLDrdLLLdlUruLLrruRu"
  "rDDrdLLLdlUruL")

(defmaze
  "#############"
  "###. * #    #"
  "# .*.*.#  $ #"
  "# ..*  #  $ #"
  "##.* ####$$ #"
  "#...#@$ $   #"
  "##.# $$  $ ##"
  "##$$   ##  ##"
  "##   ########"
  "#############")

(add-solution
  "DRRuRRUUddlldlllddllUUUrUrUrruuLLDLDlDurRdLruuurrdLulDldlluRRdddLUruul"
  "ldRurRurrdLLLDDrUluRurrddLLdlluluRRlddddddrruLdlUUUUrrurruullDurrddlUl"
  "lDulldRddddrrurrdLLLdlUUUUUluRRRRllldddddrrrrururruuluurrrddddLLrddlUU"
  "rruuLDDruuuulDDuullddRluRurrddddlldLLuRRdllLdlluRuRRdLulDrrrruLLrrUUlu"
  "urrrddddLLddrUluuuluurrrddddLruuuulllddrdDrdLLLuRdlLdLLLdlUUUUUluRRRll"
  "dddddrruRuRRdLulDrdLLLdlUUUUUluRRldddddrrrrurrrruruuuulllddrdDrdLLLLul"
  "DrdLLLdlUUUUUluRdddddrrrrurrruuuluurDDDDrdLLLLulDrdLLLdlUUUUUddddrrrru"
  "rrrruruuuLulDDDDrdLLLLulDrdLLLdlUUUUdddrrrrurrrruruuLulDDDrdLLLLulDrdL"
  "LLdlUUUddrrrrurrrruruLulDDrdLLLLulDrdLLLdlUUdrrrruruRdrruulDrdLLLLulDr"
  "dLLLdlU")

(defmaze
  "##############"
  "########   ###"
  "######   #$###"
  "#####..### ###"
  "#   # . ## ###"
  "#   .**  #  ##"
  "# $*...**   ##"
  "###  $ $.  ###"
  "######$##$####"
  "######@##  ###"
  "####### $  $ #"
  "#######      #"
  "##############")

(add-solution
  "UURULULuurDldLddRUrrdLulUUrrDLddrUllldRuruulDrdLulLulldRRRRddllUdrruul"
  "LddrUrrrdLLuuuruurrurrDDDDrdLdlLLulluurrDrDLdRluuuluurrurrdddddLLdlUrr"
  "ruuuuulldllddldddRluuuruurrurrdddddlllLUUdddlUlluullddRRRRdrUUdrrdLrur"
  "rdLDDrddlUUUUruLLLdRluLLdRRurrDDDrddlUUUUruLLLdRllluLLdRRRRuLLrrrrDDDr"
  "ddlUUUUruLLLdRurDlllluRRRdllllluRRRllluullddRRRRlluulDldRRdrruLrdrrruu"
  "LDLLrrrdrruLLLLdRurrdDDrddlUUUUruLLLrrdddddlluRdrUUUUruLLrdddrddrruLLd"
  "lUUUUruL")

(defmaze
  "###########"
  "#  ##  ####"
  "# $#@$ ####"
  "#  #$$ ####"
  "# $  $ ####"
  "#   $#....#"
  "# $$ ...# #"
  "###  ..*  #"
  "###########")

(add-solution
  "RurDDDDrdLdRluuuLLDllluurDRRRUrDDrddlllUUluRdddlUruulllddRRUdlluurDrru"
  "LdddrrrruuluLLdllldRluuurDDrruLrrrdrddlllUUddrrruululLdddrrUdlluRRlluu"
  "rrDrrrddLLLrULuRRllulldllluurDRRdddlUrRRdrUluRdllluulllddRRRRRdrUllluu"
  "llDldRRRRRlldRRuuuuuulDlDDDDuuuRurDDDDlluuRurDllllluuurDDDDldRRRRluulD"
  "ldRRldR")

(defmaze
  "############"
  "###....*.. #"
  "#####$#.*. #"
  "#  ## ###*##"
  "#    $@# . #"
  "#  $$#$$ . #"
  "# $$ $ $# ##"
  "#         ##"
  "############")

(add-solution
  "LLLrDLrurrDRRurrdLddllllllUUrurrdrRurDlllulldlddrrrrUdlllluururrdRddll"
  "lllluuRluurDldddrrrrrruulullLDrurrdrRurUULuLLLLrDDuurrrdrddldlluLrdrru"
  "rrdLuuuruLLLLLdddrdrddrrUUUUddddlluululuuurrrrrdLdddddlllUdllURRdrrruu"
  "uuuulllldddLDuruuurrrrddddddlllllllluuRRlluurDldddrrURdrrrrruuuuuullll"
  "dddlLDurruuurrrrddddddllllluUddrrrrruuuuuulllldddrDrRllulLdddrrUdrUdrr"
  "uUUUUruLLLLrrrddddddllllluRRlluurrDRRllullddrdrrrruUUUUruLLLrrddddddll"
  "UdlluRuRRllddrrruUUUUruLLrddddddllUluRRlddrruUUUUruLddddddlllllllluuRR"
  "llddrURRRRdrUluRRlddrruUUUUdddlllullDldRRRdrUluRRlddrruUUUddlldllllllu"
  "RRlluurDDldRRRRRdrUluRRlddrruUUdlllullLdlluurDDldRRRRRdrUluRRlddrruUll"
  "lullDldRRRdrUluRR")

(defmaze
  "##############"
  "#  ..........#"
  "#  ###########"
  "## #  ########"
  "##$$   #######"
  "##   $ #######"
  "####$    #####"
  "####  ##$#####"
  "#####$$  $ ###"
  "####@$   $ ###"
  "#######    ###"
  "##############")

(add-solution
  "RUUlUUllUUUluRRRRRRRRRRlllllllllddddrrRuulDLdlUUUluRRRRRRRRRlllllllldd"
  "drrrdLddrUUddddRRuLrddrrruuLLLddrUluRdlllUUluurDrrrDDldlluUddrruruullL"
  "uluurDDDrrrddldlluUlUUrurDllddrURuululDDulldRurrrddlUruLLLdlUUUluRRRRR"
  "RRRllllllldddrdRurrddlUruLLLdlUUUluRRRRRRRlllllldddrrrdddlUUruLLdlUUUl"
  "uRRRRRRllllldddrrrdddddrruLdlUUUUruLLLdlUUUluRRRRRlllldddrrrddddrrruuL"
  "LLdlUUruLLdlUUUluRRRRllldddrrrddddrrddrrruuLLLLdlUUUUruLLLdlUUUluRRRll"
  "dddrrrddddrrddrrUruLLLLdlUUUUruLLLdlUUUluRRldddrrrddddrrddrUruLLLdlUUU"
  "UruLLLdlUUUluR")

(defmaze
  "###########"
  "####  . ###"
  "####$$. $@#"
  "#   . *$###"
  "#...***   #"
  "# **$.* $ #"
  "# $.#  $  #"
  "# $.##    #"
  "###########")

(add-solution
  "LLDDrDuluLLDLLDuullddddRluuuurrRdrRuruulllDDRdrDRUrrdddllUUddrruLruulD"
  "lluRlluluurrrdDLddrrrddlllUlURuuruulllddlddRRdrURUUlLDRddlUlluuRuurrDD"
  "DuuullddRluurrdLdllddrrULulllddRRUruLrdrrULdddrUddrrruuLLrrddlUdlluRll"
  "uuurrDDLUluuurrDDLruulldRddlUdlluRdrruuullDDuurrdLrddlUrdddlUruulllddl"
  "luuRlddrruLrdRRULrruuullDDuurrdLrddlUddllluuRRDLrrruuullDDRluurrdLddrU"
  "dlddrUddrrruuLLrrddlUdlluRlluuurrDDLUluuurrDDLdlUrdddlUruulllddRRULrru"
  "ruulllDDuurrrddlLdddrUddrrruuLLrrddlUdlluRlluuurrDDLUluUluRdddrddlUlll"
  "lddrUluRRRRdrdrrruuuLLdLLLLrrrrurrdLLLLLrrddrrUruLLLddrUruL")

(defmaze
  "##############"
  "###   ## . ###"
  "###$   #.* $ #"
  "# $ #$*...## #"
  "# $ #@##*..  #"
  "#   ##   ** ##"
  "# $ $   # ####"
  "#   #     ####"
  "##############")

(add-solution
  "URuLrdRRDrDLuuuurrdLDlddrUrrruuLLullDDuurrdLrrrddlllUUdlllluullDDDDlld"
  "drrURRRdrrrUUUrrdLulDrruruulllDlllluulldddDllddrrUrrruRRRUrrruulllDlDR"
  "lullluulldddDllddrUdrUrrrdrrrUULuullluulldddDuuuurrdrdrrddrddllllulLdl"
  "luuuuRlddddrrurrrdrrruuluurDuurrrddlLLulllluulldDllddddrrurrrdrrruUlUR"
  "uLLrrurrrddlLLdddllluuRRllddrrruUlURuurrrddldLruruulllddRluLddllddrrru"
  "UruLdddllUluRRUdlddrruUruLuurrrddLdlllldlllldlluRRRRRdrUluRRlddrruUUUd"
  "dlldlllllluuurrDDuulldddrRRRRdrUlllllluuRDldRRRRRuRRRddllUluRRlddrruUU"
  "dlldllllluuurDDldRRRRdrUluRRlddrruUlldlllluuuuurrdLulDDDDldRRRRdrUluRR")

(defmaze
  "###############"
  "## .. #########"
  "## $.$ ########"
  "###.*. #  #####"
  "## * * #  ##@##"
  "## *  ##  ##$ #"
  "# *. ####  #  #"
  "# ** ####$  $ #"
  "#          $ ##"
  "####  #########"
  "###############")

(add-solution
  "DrddLdLLLUUrddrruLLdlLLLLdlUllluuRDldRRRULrUUrUruuLDLDDDLddlluuRuRldld"
  "drrrRRRRRRurrdLLLLLLLLLuUrDuUrUruulDlDDDlluuRlddrruLrrUruululDulldRDDu"
  "uurrdLrrddlULuurDrddldlUUdddlddRUUddRRRRRRurrdLLLLLLLdlUluulldRurrUrur"
  "uulDlDllDDrrUdlluurDrrUruululDulldRDDuuurrdLrrddlULuurDrddldlUllddlddR"
  "RRUUULulDrrruLrruullDDrdLuuurDluulldRRurDlddddddRRRRRRuulDrdLLLLLdlUUU"
  "UUUUruLdddddddrrrrrrurrruulDrdLLrdLLLLLLLdlUUUUUUUddddddrrrrruurDrdLLL"
  "LLLdlUUUUUUdddddrrrrruuUluurDDDDrdLLLLLdlUUUUU")

(defmaze
  "###############"
  "#   ## *...#.##"
  "# $ # $ ***..##"
  "# $ $@## *.. ##"
  "# $#$$$  ##.$ #"
  "#             #"
  "###############")

(add-solution
  "uRuRRDDRRlluuRDRdRddrruLUULrdddllllUURRRurDllllddlllURdrrrrrUdllluurrR"
  "lluullldlDururrrddlddlluRdlllllluuuurrddRluullddddrrrURRdrrrrruulluull"
  "ldlDururrrddlUruLrddrruLdlluurDrrdddllllluLuurRRdRRlluurDllllddrdrrUUd"
  "dllluuurrRdRUrrDLulldddrrrUdrruLdllllllllllluuuurrddRluullddddrrrrrrru"
  "uulllDurrrrrrrDLullldddlluRdrUUddllllURuurrRRRRllllllddRRdrUUdllluurrR"
  "RRlldddllllllluuuurrddRdrrrruulllDldRRRdrUUdllluurrRRldddllllllluuRRRd"
  "rrrruulllDlluulDldRRRdRRRdrUUUlllDldRRRdrUUdllluurrRdddllllllUluRRRdrr"
  "rruulllDldRRRdrUUddrrrUdrruL")

(defmaze
  "###########"
  "########  #"
  "####   #$ #"
  "###     $ #"
  "#   #$##$ #"
  "# $  *.#@ #"
  "# $#$...# #"
  "#    #.$# #"
  "#  $$#..$ #"
  "#### ....##"
  "###########")

(add-solution
  "rdddLdlluuuuLrddddrruruuuuuLLLulldRDDrddddrruruuuuullLulldlddRlulldddr"
  "rrURUrDlldllluuurrururrdrrrdddddldlluRdlllULulluuurrdRdrrrDulllululldd"
  "drrdrdrruUruLdddrUdlllululluuurrururrdrrrdddddLruuuuulllulldRRllldlldd"
  "drrdrdrruuuuLUdrddddllululluuurrdRdRlululldddrrdrUllluuRDuurururrdLdDr"
  "DullLullddrUdlddRUrrUdllluurruuRurrdLddLLuuRurDllddrddddrrrruruuuuulLL"
  "ulldlddrrrdLulluururrdrrrdddddldlllluuUdddrrrruruuuuulllulldlddRddddrr"
  "uRdllluuurrrDullldddrruUruLdddrUdllluullluuRDlddrRuruuLuururrdrrrddddd"
  "LdllllUllluRluuuRDDlddrrrdrruuuuLLLrrrddddllullluurUrrdRurDllulldlddrr"
  "rdrrrruruuuuulllulldlDururrdrrrdddddldllllullluuruRRdrrDDuullulldlddrr"
  "rdrRRluuuuLrddddllullluururuururrdLDDrddddllullluururRdRluluuRurrdLdDr"
  "DDDRdLuuuulldRurDDDuulluluuRurDDDldRurDDurDDululuuurrrrdddddLruuulUruL"
  "LLulDDDllldlddrrrUURurDluuullddRluurrdDldRRllddllUUluRRRluurrdDlllddRd"
  "rUURldllluuuRdrRRuurrrruulDrdLLLulDDDuullDldRR")

(defmaze
  "#############"
  "#######   ###"
  "###### $#   #"
  "####  $ ### #"
  "# . $$@$ #  #"
  "# **.#$$    #"
  "##$..# $ #  #"
  "##.*.  # $ ##"
  "##....###  ##"
  "#############")

(add-solution
  "DRRddRdrUUruuuullullDlDllDDDDuuulDulldRDDRluuurrdLrdLddRUUluulldRurrur"
  "rururrdrrddddldlluullUdrruLULrddlULLLDDrdrrUUruLrrdddrruruuuullullDDul"
  "DruurrdrrddddldlluullddlluluuruRRlldlddrddlUUrdrruurrddrruruuuullulldl"
  "dlldDDldRuuuurrururrdrrddddldlluullULrdrrddrruruuuullulldlDllDlDlluRRd"
  "DrddlUlUdrruLuuRurrururrdrrddddldlluullddLLdlUUrdrruurrddrruruuuullull"
  "dldlldDDDrruuULulDDDuurrdddlLulDruuulDulldRurrdddllUdrrrruurruLLLulDDD"
  "uurrdddlLLrrruurrddrruruuuullulldDuurrdrrddlDLLuLLLulDDDuurrdddlLrruuu"
  "rrddLUruLLLLdddrrUUruLLulDDDuurrruulDrdLLulDDurrrrdddrrUruLLLdlUruLLLu"
  "lDL")

(defmaze
  "#############"
  "####   .##  #"
  "#@$ . #*##$ #"
  "##$*.*..#   #"
  "#  #**$     #"
  "#  ....* #$ #"
  "## $$ *.##  #"
  "##  $ #  $$ #"
  "##  ####    #"
  "#############")

(add-solution
  "RDDldRRDuRRdLuUURDRRDulllddrURuLrrRdLruRurrdddddlllulULUdrdrdrrruuuuul"
  "ldllDlluuuurrDDDrdLuLruuullddRdrrrurrdddddlllulUlUdrdrdrrruuuuulldllDl"
  "lldRlulluuRRRuurrdDDrdLLLUUlllddrRUruLdddrUrruuuulllDDrRDrrrurrdddddll"
  "lulULUdrdrdrrruuuuulldllDllllluuRRRDuluurrrdDDLdLLUUrDrrrdLrururrddddd"
  "lllulUdrdrrruuuuulldlluuullDDRlldddRRUULullldddRRRUrurrdLrururrdddddll"
  "lulULUdrdrdrrruuuuulldllDllldlluuurrruurrdDDrdLrururrdddddlllulUlLuUUl"
  "DrddrrdrdrrruuuuulldllDLruuLDlluRllldddrRUUruLdddRURurrdLuuuulllDDrRdr"
  "rrurrdddddlllulULUdrdrdrrruuuuulldllDllldlluuuRRDuuurrrdDlLDlUlldddrrU"
  "UrurrDrdLrururrdddddlllulUdrdrrruuuuulldllLulLdddRRUULulDrrruuullDulDr"
  "urrdddrrurrdLLLdLUUUddrrrrddlUruLLLdlUUdrrrrddddlUUUruLLLdlUrrrddddllu"
  "RdrUUUruLLLLLLrrrrrruuulDDrdLLLLLrrrrdddlllUlllllddrUluRRRRldLullddrUl"
  "uRRRllluuuuR")

(defmaze
  "#############"
  "### @ #######"
  "### $$#######"
  "#  ## $. ####"
  "# $## *. ####"
  "#  . ..*  ###"
  "###**..* $ ##"
  "##  .$$# $  #"
  "## $.   #   #"
  "##  #   #####"
  "#############")

(add-solution
  "rDDRDrrddLLDurruullDurrddlUlllDDRUdddrruLdlUUluuRDuRRdLullddllddrUluRd"
  "RRdrruLUULullDRlDlddrUrURldlluRuurrDLulDrrrruLLruulDrdrdrdRdrruLuLuLLd"
  "RuuuLDlluRdrrddlllUdrrruullDurrddlUlldlluRRRuulDrdLdllDlddrUrRUUlDulDr"
  "ruUrRdrdRdrruLuLuLLdLLuRuulDrddrRuuuLrdddlluuRldRdlldddrruLUUruullDRdd"
  "llluuRRDrdLrddlUruuluurDDlullddlddrURRUULuRdddlUdlluRdrruuullDDuurrdLD"
  "rurrrdRdrruLuLLLLLuurrDullddrUdRRdRdrruLuLuLLuurDlddlluRllllluurDldRRR"
  "RddlUruuuRluulldRurDDDDLddrUrrRdRdrruLuLLLLLdddrruLUluRdddlUUlllddrUlu"
  "RdRRdrruLLruuRRdRdrruLuLLLUdrrrddlUruLLruL")

(defmaze
  "###########"
  "###  ######"
  "#   $######"
  "#   $ #####"
  "# $##  $@##"
  "# $...*$###"
  "#   #** ###"
  "### *.*.* #"
  "####* ..  #"
  "#### $ *  #"
  "####    ###"
  "###########")

(add-solution
  "LDDDuuLLDDRdRUlluuuuLLullddddrrdRDRUrrdLulllulluuuurrdrrddrrDulluullul"
  "lddddrrdrrUdrdddllURUruUlDllulluuuurrdrrdDuullullddddrrdrrruruuLrddldd"
  "rUlllDRddrrULrrruuLLLDurrrddllUluRluUrDlddlluRdrrddlUdllURuulluuRRDrrD"
  "DLULuullddRDRddrrUrruuLLuuLLdDDlululluuuurrdrrDDrrddrrddlldlllUUruRRuu"
  "lluullullddddrrdRdddrrruULuUddrddllluuuRUrrDDlUlldddrrrUrruuLDuLLdRddl"
  "lluuuRllulluuuurrdrrdDDuuullullddddrrdrdddrrruuuuuuLrddddddllluuulullu"
  "uuurrdrrDrrddddlLrrddllluUddrrULdlUrrruurrddLUluuuullDRurDDDuulluullul"
  "lddddrruRRRurDDulllldlluRRlluuRDlddrruRRuulLuurDlllddddrrurruuLullDldd"
  "RRRllluurDldRRluuurrdLulDDldR")

(defmaze
  "############"
  "####  ######"
  "#### $$  $ #"
  "###@$  $   #"
  "# $$## . ###"
  "#  ..*.*####"
  "# $ #*...###"
  "# # *. #####"
  "#  # * #####"
  "##    ######"
  "############")

(add-solution
  "RRluurDRDDDLDRddLUruulLLddRRUruuuulldlDDrRdrUUrruuLLrrddllUdrruulDlddl"
  "lluurRluurDRDDurrdLululldlddrrdrRUUruuLLrrddlUdddlluRdrUUruuLDlDDuurrd"
  "LdllddddllluluuuRRuurRRlllddlldddrdrrruuuurrdLruuruulDllllddRRDrruUruL"
  "dddlluRllllldddrdrrruUllUdrrddllluluuRuRRRDDllUdrruulLrrdrrULLddllulld"
  "drdrrrUrULuurDruUruulDLDDuurrdLdllddddllluluurruuurRRllldddllddrdrrruu"
  "uurrdLruuruulDllllddRRDrruUruLdddlluRdrUlllldllddrdrrrurUdldllluluurru"
  "rrdDuurrdLruUruulDLDDlllddRlullddrdrrUdlluluuuRRRRRuurrdLDlllluurRRlll"
  "ddlluRdrrrrruruulDlDurrdLdLLrruullllDuruurDldRRDrdLruruulDDuuLdlluurDl"
  "dRRurDrrruLLLLdlluurDldRurDD")

(defmaze
  "##############"
  "####   ###  ##"
  "# $ $  ##   ##"
  "#.  $####$  ##"
  "#.*  $   $ ###"
  "#*.#$*.$ #$$ #"
  "#......* # @ #"
  "##    ########"
  "##############")

(add-solution
  "lUULLddLLLLdlluuUruuRlddldddrrruruuLLDRdrrruuLDuLDlluRRdlddllluuuRRluu"
  "rDldlluuRDrruLdlldRdddrrruulULulDrrdrruLLLuurDldRlulldRurrddrddllluUUl"
  "uRddddrrruuluLrdrddllluuUluuRDDrruLruurrdLulDLddlluuRRddrdrruLLLuuRDld"
  "RluulldRurrdddrddlUdlluRRlllUURDuuurrddLrDrruLLuulldDlddRdrrUdrUdllluR"
  "RllluurDuuurrddLrrrdrruLLLLuulldDlddRdrruURRllddrUdllluRRllluurDuuurrd"
  "DLruulldDlddRUUdddrruLdlUluurrrDurrdLullllddrrdrrULdlluluurrrrDulllldd"
  "rdrruUruLdddrUdllluRRllluurDldRdrruuuuullDDDuuurrurrdLulDDDDuLruulllDD"
  "DuurDrrrrrrdLrurrUruulDlDurrdLdLLLLLLLLulDrrrrrrrrruulDrdddrruLdlUULLL"
  "LLLLLrrrrrrruurrdLulDrdLLLLLLLruLL")

(defmaze
  "###########"
  "#####  ####"
  "####. .  ##"
  "###..**# ##"
  "###* *.  ##"
  "#  . ##$###"
  "# $.$##   #"
  "##$$ $@$  #"
  "##    #   #"
  "###########")

(add-solution
  "RlLdllURRlUUlDulldRurrdddllURlURurUUlDDDlddrUdrruLUUdddlluuRlddrruLuur"
  "DluuurDDldlluRdrddlUrdrruLLuUruulDDruRRULuurDrrddLLUdrruulLddrDDrrddlU"
  "dlULLdlluRRRRlllUUlDulldRurrdddllUURlddrruLuurDluuuRDRRuurrddLLLLDlddd"
  "rrurrUUddlldlluuururruurrddLdddlldlluuuruuRdldldddrrurruuuruullulDrrrd"
  "dlLUdrdddlldlluuuruuRlddldddrrurruuulLullDDrUluRdrrrdddlldlluUddrruLUU"
  "UrrrddddrruuLrdLLLLdlUUlUUrrrrdDuullllddrUluRRldddrrRdrruLLLLdlUUUluRd"
  "dllluRdrUrdddlUdlUUluR")

(defmaze
  "###########"
  "########  #"
  "#######   #"
  "#  ####$$ #"
  "# $$ # . ##"
  "# $ .*.. ##"
  "#  $ .# ###"
  "#  # ..$###"
  "## *.*. ###"
  "## *$$*####"
  "###   @####"
  "###########")

(add-solution
  "llUULrddrrUUrUdldLdllUlUUluuRDRRuLrdDrDrddlUdlluRlluUluRRurRRurrUdlldl"
  "DLdDLddrrruuLLrrddlUruLUluurDDDrddlUdlluRUUdddrruuLrddlluRuulDruuulDDr"
  "dddllulUUluuRDluuurDlddrRRuuLDrddrdddlluluUluuRRurDllluRdlddrUUddddrdr"
  "ruLruulDLrrddllUlUUluurrrDDuurDDDrruLdlUluulllddrddrdrrUdlluRUUdddrrUU"
  "LrddllluluuluurrrrDDlDLddrrruuLLrrddllURuulDruuulDDrdrruLdlddrUdllluRR"
  "uUlDuuulllddrURurDllluRluurDlddrrruuLDrdDrdddllulUUluuRRurDllluRdlddrU"
  "UddddrdrruLruulDLrrddllUlUUluurrrrrurrdLLLLDDrrrUdllluurDldRdddlluluul"
  "uurDluuurDlddrRRuuLDrddrdddlluluUluuRRurDllluRdlddrUUddddrdrruLruulDLr"
  "rddllUlUUluurrrrrurrdLLrruuruulDlDDuurrdLulDrddLdddLLuRdrUUddllddrUUdd"
  "llluRRuulDrddlluluRRRdLulluluurrrDDuulllddrdrRdrUllluURurDDullluRluurD"
  "DldRddrRuuuuLrddddddlUluuuuRurDDDDuuulldRurDDullluuurDldRRdldddrdrruL")

(defmaze
  "##############"
  "#  ####.##   #"
  "#   ## ..# $ #"
  "# $ ##$.$##  #"
  "#  $ #....   #"
  "# $ $ #.*#####"
  "# #   $*.. $@#"
  "# $$ # .$ ####"
  "#    #  ######"
  "##############")

(add-solution
  "LLLdLLULLuLuLuullddddddrrruUlUUddrddllluuuuRUldddddrrruuruLuLrddlUdrrr"
  "rdrruLLUUrDluUUrDDldddrruLdlUUUrDlddrruLdlULLulldRRRRddlUruUrDuuuullDR"
  "urDDulDDDLddrUUUUrrrrruuulDDrdLLLLuulldRurDlDDDlddrUUUURuulldRurDldddl"
  "LulldRRRdrUUUUluRddddlllddlUdlluRRdrUllluuRRDRRRdrUUUUdddllllulluurDld"
  "ddrRdrUluRRRdrUUUddlluLuLDDrddlluRdrUluRRRdrUUdlllldlluuRRurDldRRRdrUr"
  "rdLulllllulluurDldRRurDldRRRRRllllluulluuurDDDldRRurDldRRRRlllluuluurD"
  "DldRurDldRRR")

(defmaze
  "##############"
  "## @  #   ####"
  "# $$$$#. *.###"
  "#  # *..$*$. #"
  "# $# $**.... #"
  "#     .#$.$$ #"
  "##   ###  #  #"
  "##############")

(add-solution
  "rDDDuRRDRRULdrRRULdlluuurrDDLruulldRddlUdlLuRlluurDldRddLUUdddlluRlluu"
  "rDldRdrrULrruulDruuulDullDRRlllddrUluRurrdDrddLUUdrrrrurrdLLurrrrdLrdd"
  "lUULulldRdRluurDrruLLdllLLLdldllUUluururrrDDlDRuuullldldddrdrrULrruulD"
  "ruuulDulldRRlllddrUluRurrdDrddLUUdrRRRRddlUruRurrdLrddlULrruulDulDlLuR"
  "RlldddrUUrrrddlULullLLLdldllUUluururrrDDlDRRRRRlllluuullldldddrdrrULrr"
  "uulDruuulDulldRRlllddrUluRurrdDrddLdllUUluurRurrDDuulldRllldddrdrruUUd"
  "RRRRllllddlluuluurrurrDDldRRRlluuulldRurDDldRRllddlluRdrUUUruulldRurDD"
  "ldRdllluUluRRRurD")

(defmaze
  "###########"
  "#     . ###"
  "#..*.**$###"
  "# $#..# ###"
  "# $ .*  ###"
  "#$#$$   $ #"
  "#@#       #"
  "###########")

(add-solution
  "UURRlUURurrrrDDDlddlllURdrruuruuulllldlluRdddlUruuRRRDLdDLddrrruuLUluu"
  "llddldRuuurrddrdrddlULrruruuuLLDldDRluurDDuuurrdLulDlLullDRRRRurrdddLd"
  "dllUUUdddrruuruuullLLdlddRluururrrrdddlddrUdrruLLdllluUlluururrrrddDlL"
  "LddrrURuuuulldDuulldRurrrdddddllluRluUrDlddrrruuuuulllldlddRluururrDLu"
  "rrrdddddllluUUdddrrruuLLUlDrrruuulllDurrrdddllllluuRurrrrdddddllllUdrr"
  "rruuuuulllldlluRRldddRddrrrruuuuullDLLulDrrrurrdddddlllluuRUdlddrrrruu"
  "uuulldLLulldRurRRldddlddrUUUdddrrruulLUdrrddllUUdlUddrrUruLddrrruLLLLd"
  "lUlulUdlU")

;;;

(setf *mazes*     (nreverse *mazes*)
      *solutions* (nreverse *solutions*))
